//----------------------------------------------------------------------------------------
/**
 * \file    data.h
 * \author  Zuzana Slaninova
 * \using   asteroids
 * \date    2023
 * \brief   Basic defines and data structures.
 */
//----------------------------------------------------------------------------------------

#ifndef __DATA_H
#define __DATA_H

#define WINDOW_WIDTH   750
#define WINDOW_HEIGHT  750
#define WINDOW_TITLE   "Pyramids World"

// keys used in the key map
enum { KEY_LEFT_ARROW, KEY_RIGHT_ARROW, KEY_UP_ARROW, KEY_DOWN_ARROW, KEY_SPACE, KEYS_COUNT };

#define DUCKS_COUNT_MIN  5
#define DUCKS_COUNT_MAX 10


#define DUCK_SIZE        0.05f

#define BALON_VIEW_ANGLE_DELTA 2.0f // in degrees


#define BALON_SIZE   0.1f
#define BALON_SPEED_INCREMENT  0.025f
#define BALON_SPEED_MAX        1.0f

#define DUCK_SPEED_MAX         0.5f

#define BILLBOARD_SIZE   0.1f
#define FEATHER_BILLBOARD_SIZE   0.08f

#define SCENE_WIDTH  1.0f
#define SCENE_HEIGHT 1.0f
#define SCENE_DEPTH  1.0f

#define CAMERA_ELEVATION_MAX 45.0f
#define CAMERA_SPEED 5.0f

/// Carpet constants
#define CARPET_SHAPE_RATIO  3.0f       //how many times the longer side is longer then the shorter side
#define CARPET_WAVE_COUNT   2          //count of waves on longer side
#define CARPET_WAVE_HEIGHT  1.0f       //relative: 1.0 .. 1/12 of lenght
#define CARPET_SLICE_COUNT  120        //count of slices on longer side - for generating a mesh

//enum for camera modes
enum CameraEnum
{
    TOP,
    FREE,
    CURVE,
    SIDE
};

//enum for stencil
enum StencilIDs{
    ZERO,
    PYRAMID,
    DUCK,
    ROCK,
    ALOE
};

// default shaders - color per vertex and matrix multiplication
const std::string colorVertexShaderSrc(
    "#version 140\n"
    "uniform mat4 PVMmatrix;\n"
    "in vec3 position;\n"
    "in vec3 color;\n"
    "smooth out vec4 theColor;\n"
    "void main() {\n"
	"  gl_Position = PVMmatrix * vec4(position, 1.0);\n"
	"  theColor = vec4(color, 1.0);\n"
    "}\n"
);

const std::string colorFragmentShaderSrc(
    "#version 140\n"
    "smooth in vec4 theColor;\n"
    "out vec4 outputColor;\n"
    "void main() {\n"
    "  outputColor = theColor;\n"
    "}\n"
);

// each vertex shader receives screen space coordinates and calculates world direction
const std::string skyboxFarPlaneVertexShaderSrc(
  "#version 140\n"
  "\n"
  "uniform mat4 inversePVmatrix;\n"
  "in vec2 screenCoord;\n"
  "out vec3 texCoord_v;\n"
  "\n"
  "void main() {\n"
  "  vec4 farplaneCoord = vec4(screenCoord, 0.9999, 1.0);\n"
  "  vec4 worldViewCoord = inversePVmatrix * farplaneCoord;\n"
  "  texCoord_v = worldViewCoord.xyz / worldViewCoord.w;\n"
  "  gl_Position = farplaneCoord;\n"
  "}\n"
);

// fragment shader uses interpolated 3D tex coords to sample cube map
const std::string skyboxFarPlaneFragmentShaderSrc(
  "#version 140\n"
  "\n"
  "uniform samplerCube skyboxSampler;\n"
  "in vec3 texCoord_v;\n"
  "out vec4 color_f;\n"
  "\n"
  "void main() {\n"
  "  color_f = texture(skyboxSampler, texCoord_v);\n"
  "}\n"
);


//
// sand billboard geometry definition 
//

const int sandNumQuadVertices = 4;
const float sandVertexData[sandNumQuadVertices * 5] = {

    // x      y     z     u     v
    -1.0f, -1.0f, 0.0f, 0.0f, 0.0f,
     1.0f, -1.0f, 0.0f, 1.0f, 0.0f,
    -1.0f,  1.0f, 0.0f, 0.0f, 1.0f,
     1.0f,  1.0f, 0.0f, 1.0f, 1.0f,
};

//
// feather billboard geometry definition 
//
const int featherNumQuadVertices = 4;
const float featherVertices[] = {
    // x      y     z     u     v
    -1.0f, -1.0f, 0.0f, 0.0f, 0.0f,
     1.0f, -1.0f, 0.0f, 1.0f, 0.0f,
    -1.0f,  1.0f, 0.0f, 0.0f, 1.0f,
     1.0f,  1.0f, 0.0f, 1.0f, 1.0f,
};

//
// feather billboard geometry definition 
//
const int aloeNumQuadVertices = 4;
const float aloeVertices[] = {
    // x      y     z     u     v
    -1.0f, -1.0f, 0.0f, 0.0f, 0.0f,
     1.0f, -1.0f, 0.0f, 1.0f, 0.0f,
    -1.0f,  1.0f, 0.0f, 0.0f, 1.0f,
     1.0f,  1.0f, 0.0f, 1.0f, 1.0f,
};

//
// moving texture geometry definition 
//
const int texture1NumQuadVertices = 4;
const float texture1VertexData[texture1NumQuadVertices * 5] = {

    // x      y      z     u     v
    -1.0f,  0.15f, 0.0f, 0.0f, 1.0f,
    -1.0f, -0.15f, 0.0f, 0.0f, 0.0f,
     1.0f,  0.15f, 0.0f, 2.0f, 1.0f,
     1.0f, -0.15f, 0.0f, 2.0f, 0.0f
};


//
// rock geometry definition
//

const int rockNAttribsPerVertex = 3;
const int rockNTriangles = 36;

const float rockVertices[] = {
    // x            y           z
    - 0.150527f, 1.500827f, 0.013990f,
    1.000000f, - 1.000000f, - 1.000000f,
    - 0.217775f, 1.448054f, 0.208883f,
    1.000000f, - 1.000000f, 1.000000f,
    - 0.473116f, 1.438034f, 0.127285f,
    - 1.000000f, - 1.000000f, - 1.000000f,
    - 0.425257f, 1.387842f, 0.251082f,
    - 1.000000f, - 1.000000f, 1.000000f,
    0.490598f, - 0.115575f, 1.284563f,
    - 0.839422f, - 0.038892f, 0.767983f,
    1.240586f, 0.154871f, - 0.721586f,
    - 0.726168f, - 0.002012f, - 0.651267f,
    0.758708f, 0.708623f, 0.682193f,
    - 0.577905f, 0.668806f, 0.489843f,
    0.533435f, 0.986544f, - 0.408007f,
    - 0.766408f, 0.706299f, - 0.259474f,
    - 0.060598f, 0.999673f, 0.456307f,
    - 0.551626f, 1.213879f, 0.195951f,
    0.077739f, 1.413458f, - 0.192866f,
    - 0.552618f, 1.278288f, 0.025488f,
};

const unsigned int rockIndices[] = {
    5 - 1, 3 - 1, 1 - 1,
    9 - 1, 8 - 1, 4 - 1,
    8 - 1, 12 - 1, 6 - 1,
    2 - 1, 8 - 1, 6 - 1,
    11 - 1, 4 - 1, 2 - 1,
    12 - 1, 2 - 1, 6 - 1,
    14 - 1, 9 - 1, 13 - 1,
    13 - 1, 11 - 1, 15 - 1,
    16 - 1, 10 - 1, 14 - 1,
    15 - 1, 12 - 1, 16 - 1,
    17 - 1, 14 - 1, 13 - 1,
    17 - 1, 15 - 1, 19 - 1,
    18 - 1, 16 - 1, 14 - 1,
    19 - 1, 16 - 1, 20 - 1,
    7 - 1, 17 - 1, 3 - 1,
    3 - 1, 19 - 1, 1 - 1,
    5 - 1, 18 - 1, 7 - 1,
    1 - 1, 20 - 1, 5 - 1,
    5 - 1, 7 - 1, 3 - 1,
    9 - 1, 10 - 1, 8 - 1,
    8 - 1, 10 - 1, 12 - 1,
    2 - 1, 4 - 1, 8 - 1,
    11 - 1, 9 - 1, 4 - 1,
    12 - 1, 11 - 1, 2 - 1,
    14 - 1, 10 - 1, 9 - 1,
    13 - 1, 9 - 1, 11 - 1,
    16 - 1, 12 - 1, 10 - 1,
    15 - 1, 11 - 1, 12 - 1,
    17 - 1, 18 - 1, 14 - 1,
    17 - 1, 13 - 1, 15 - 1,
    18 - 1, 20 - 1, 16 - 1,
    19 - 1, 15 - 1, 16 - 1,
    7 - 1, 18 - 1, 17 - 1,
    3 - 1, 17 - 1, 19 - 1,
    5 - 1, 20 - 1, 18 - 1,
    1 - 1, 19 - 1, 20 - 1,
};




#endif // __DATA_H
