//----------------------------------------------------------------------------------------
/**
 * \file    render_stuff.cpp
 * \author  Zuzana Slaninova
 * \using   asteroids
 * \date    2023
 * \brief   Rendering stuff - drawing functions for models, etc..
 */
//----------------------------------------------------------------------------------------

#include <iostream>
#include "pgr.h"
#include "render_stuff.h"
#include "data.h"
#include "spline.h"
#include <vector>

MeshGeometry* duckGeometry = NULL;
MeshGeometry* balonGeometry = NULL;
MeshGeometry* skyboxGeometry = NULL;
MeshGeometry* pyramidGeometry = NULL;
MeshGeometry* pyramidGeometry2 = NULL;
MeshGeometry* pyramidGeometry3 = NULL;
MeshGeometry* flagGeometry = NULL;
MeshGeometry* aloeGeometry = NULL;
MeshGeometry* sandGeometry = NULL;
MeshGeometry* featherGeometry = NULL;
MeshGeometry* textureGeometry = NULL;
MeshGeometry* rockGeometry = NULL;
MeshGeometry* aloeAGeometry = NULL;
MeshGeometry* carpetGeometry = NULL;

typedef std::vector<MeshGeometry*> FlagGeometryList;

FlagGeometryList flagGeometryList;

const char* DUCK_MODEL_NAME = "data/duck.obj";
const char* PYRAMID_1_MODEL_NAME = "data/pyramid1.obj";
const char* PYRAMID_2_MODEL_NAME = "data/pyramid2.obj";
const char* PYRAMID_3_MODEL_NAME = "data/pyramid3.obj";
const char* ALOE_MODEL_NAME = "data/aloe.obj";
const char* BALON_MODEL_NAME = "data/balon.obj";
const char* FLAG_MODEL_NAME_2 = "data/flag2.obj";
const char* FLAG_MODEL_NAME_3 = "data/flag3.obj";
const char* FLAG_MODEL_NAME_4 = "data/flag4.obj";
const char* FLAG_MODEL_NAME_5 = "data/flag5.obj";
const char* FLAG_MODEL_NAME_6 = "data/flag6.obj";
const char* SKYBOX_CUBE_TEXTURE_FILE_PREFIX = "data/skybox";
const char* SAND_TEXTURE_NAME = "data/sand.png";
const char* FEATHER_TEXTURE_NAME = "data/feather2.png";
const char* MOVING_TEXTURE_NAME = "data/snake.png";
const char* ALOE_A_TEXTURE_NAME = "data/aloe_a.png";
const char* CARPET_TEXTURE_NAME = "data/carpet2.png";

SCommonShaderProgram shaderProgram;
SCommonShaderProgram carpetShaderProgram;

bool useLighting = false;
bool fog = false;

//carpet geometry
float carpetVertexData[(CARPET_SLICE_COUNT + 1) * 2 * (3 + 2)];

struct SkyboxFarPlaneShaderProgram {
  // identifier for the shader program
  GLuint program;                 // = 0;
  // vertex attributes locations
  GLint screenCoordLocation;      // = -1;
  // uniforms locations
  GLint inversePVmatrixLocation; // = -1;
  GLint skyboxSamplerLocation;    // = -1;
} skyboxFarPlaneShaderProgram;

struct SandShaderProgram {
    // identifier for the shader program
    GLuint program;              // = 0;
    // vertex attributes locations
    GLint posLocation;           // = -1;
    GLint texCoordLocation;      // = -1;
    // uniforms locations
    GLint PVMmatrixLocation;     // = -1;
    GLint VmatrixLocation;       // = -1;
    GLint timeLocation;          // = -1;
    GLint texSamplerLocation;    // = -1;
    GLint frameDurationLocation; // = -1;

} sandShaderProgram;

struct FeatherShaderProgram {
    // identifier for the shader program
    GLuint program;              // = 0;
    // vertex attributes locations
    GLint posLocation;           // = -1;
    GLint texCoordLocation;      // = -1;
    // uniforms locations
    GLint PVMmatrixLocation;     // = -1;
    GLint VmatrixLocation;       // = -1;
    GLint timeLocation;          // = -1;
    GLint texSamplerLocation;    // = -1;
    GLint frameDurationLocation; // = -1;

} featherShaderProgram;

struct AloeShaderProgram {
    // identifier for the shader program
    GLuint program;              // = 0;
    // vertex attributes locations
    GLint posLocation;           // = -1;
    GLint texCoordLocation;      // = -1;
    // uniforms locations
    GLint PVMmatrixLocation;     // = -1;
    GLint VmatrixLocation;       // = -1;
    GLint timeLocation;          // = -1;
    GLint texSamplerLocation;    // = -1;
    GLint frameDurationLocation; // = -1;

} aloeAShaderProgram;

struct Texture1ShaderProgram {
    // identifier for the shader program
    GLuint program;           // = 0;
    // vertex attributes locations
    GLint posLocation;        // = -1;
    GLint texCoordLocation;   // = -1;
    // uniforms locations
    GLint PVMmatrixLocation;  // = -1;
    GLint timeLocation;       // = -1;
    GLint texSamplerLocation; // = -1;
    bool fog;
} texture1ShaderProgram;

void generateCarpetVertexData() {
    // carpetVertexData is the global var
    float x, u;
    const float y0 = 0, y1 = 1, z = 0, v0 = 0, v1 = 1;
    int i = 0;
    int idx = 0;

    while (i <= CARPET_SLICE_COUNT) { //value i == CarpetSliceCount is included for last edge
        x = CARPET_SHAPE_RATIO * i / CARPET_SLICE_COUNT;
        u = i / CARPET_SLICE_COUNT;
        carpetVertexData[idx] = x;
        idx++;
        carpetVertexData[idx] = y1;
        idx++;
        carpetVertexData[idx] = z;
        idx++;
        carpetVertexData[idx] = u;
        idx++;
        carpetVertexData[idx] = v1;
        idx++;
        carpetVertexData[idx] = x;
        idx++;
        carpetVertexData[idx] = y0;
        idx++;
        carpetVertexData[idx] = z;
        idx++;
        carpetVertexData[idx] = u;
        idx++;
        carpetVertexData[idx] = v0;
        idx++;
        i++;
    }
}

void setTransformUniforms(const glm::mat4 &modelMatrix, const glm::mat4 &viewMatrix, const glm::mat4 &projectionMatrix) {

  glm::mat4 PVM = projectionMatrix * viewMatrix * modelMatrix;
  glUniformMatrix4fv(shaderProgram.PVMmatrixLocation, 1, GL_FALSE, glm::value_ptr(PVM));

  glUniformMatrix4fv(shaderProgram.VmatrixLocation, 1, GL_FALSE, glm::value_ptr(viewMatrix));
  glUniformMatrix4fv(shaderProgram.MmatrixLocation, 1, GL_FALSE, glm::value_ptr(modelMatrix));

  // just take 3x3 rotation part of the modelMatrix
  // we presume the last row contains 0,0,0,1
  const glm::mat4 modelRotationMatrix = glm::mat4(
    modelMatrix[0],
    modelMatrix[1],
    modelMatrix[2],
    glm::vec4(0.0f, 0.0f, 0.0f, 1.0f)
  );
  glm::mat4 normalMatrix = glm::transpose(glm::inverse(modelRotationMatrix));

  //or an alternative single-line method: 
  //glm::mat4 normalMatrix = glm::transpose(glm::inverse(glm::mat4(glm::mat3(modelRotationMatrix))));

  glUniformMatrix4fv(shaderProgram.normalMatrixLocation, 1, GL_FALSE, glm::value_ptr(normalMatrix));  // correct matrix for non-rigid transform
}

void setTransformUniformsCarpet(const glm::mat4& modelMatrix, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {

    glm::mat4 PVM = projectionMatrix * viewMatrix * modelMatrix;
    glUniformMatrix4fv(carpetShaderProgram.PVMmatrixLocation, 1, GL_FALSE, glm::value_ptr(PVM));

    glUniformMatrix4fv(carpetShaderProgram.VmatrixLocation, 1, GL_FALSE, glm::value_ptr(viewMatrix));
    glUniformMatrix4fv(carpetShaderProgram.MmatrixLocation, 1, GL_FALSE, glm::value_ptr(modelMatrix));

    // just take 3x3 rotation part of the modelMatrix
    // we presume the last row contains 0,0,0,1
    const glm::mat4 modelRotationMatrix = glm::mat4(
        modelMatrix[0],
        modelMatrix[1],
        modelMatrix[2],
        glm::vec4(0.0f, 0.0f, 0.0f, 1.0f)
    );
    glm::mat4 normalMatrix = glm::transpose(glm::inverse(modelRotationMatrix));

    //or an alternative single-line method: 
    //glm::mat4 normalMatrix = glm::transpose(glm::inverse(glm::mat4(glm::mat3(modelRotationMatrix))));

    glUniformMatrix4fv(carpetShaderProgram.normalMatrixLocation, 1, GL_FALSE, glm::value_ptr(normalMatrix));  // correct matrix for non-rigid transform
}

void setMaterialUniforms(const glm::vec3 &ambient, const glm::vec3 &diffuse, const glm::vec3 &specular, float shininess, GLuint texture) {

  glUniform3fv(shaderProgram.diffuseLocation,  1, glm::value_ptr(diffuse));  // 2nd parameter must be 1 - it declares number of vectors in the vector array
  glUniform3fv(shaderProgram.ambientLocation,  1, glm::value_ptr(ambient));
  glUniform3fv(shaderProgram.specularLocation, 1, glm::value_ptr(specular));
  glUniform1f(shaderProgram.shininessLocation,    shininess);

  if(texture != 0) {
    glUniform1i(shaderProgram.useTextureLocation, 1);  // do texture sampling
    glUniform1i(shaderProgram.texSamplerLocation, 0);  // texturing unit 0 -> samplerID   [for the GPU linker]
    glActiveTexture(GL_TEXTURE0 + 0);                  // texturing unit 0 -> to be bound [for OpenGL BindTexture]
    glBindTexture(GL_TEXTURE_2D, texture);
  }
  else {
    glUniform1i(shaderProgram.useTextureLocation, 0);  // do not sample the texture
  }

  //glUniform1i(shaderProgram.fog, fog);

  GLuint location = glGetUniformLocation(shaderProgram.program, "isFog");
  glUniform1f(location, fog);

}

void setMaterialUniformsCarpet(const glm::vec3& ambient, const glm::vec3& diffuse, const glm::vec3& specular, float shininess, GLuint texture) {

    glUniform3fv(carpetShaderProgram.diffuseLocation, 1, glm::value_ptr(diffuse));  // 2nd parameter must be 1 - it declares number of vectors in the vector array
    glUniform3fv(carpetShaderProgram.ambientLocation, 1, glm::value_ptr(ambient));
    glUniform3fv(carpetShaderProgram.specularLocation, 1, glm::value_ptr(specular));
    glUniform1f(carpetShaderProgram.shininessLocation, shininess);

    if (texture != 0) {
        glUniform1i(carpetShaderProgram.useTextureLocation, 1);  // do texture sampling
        glUniform1i(carpetShaderProgram.texSamplerLocation, 0);  // texturing unit 0 -> samplerID   [for the GPU linker]
        glActiveTexture(GL_TEXTURE0 + 0);                  // texturing unit 0 -> to be bound [for OpenGL BindTexture]
        glBindTexture(GL_TEXTURE_2D, texture);
    }
    else {
        glUniform1i(carpetShaderProgram.useTextureLocation, 0);  // do not sample the texture
    }

    //glUniform1i(shaderProgram.fog, fog);

    GLuint location = glGetUniformLocation(carpetShaderProgram.program, "isFog");
    glUniform1f(location, fog);

}

void drawBalon(BalonObject *balon, const glm::mat4 & viewMatrix, const glm::mat4 & projectionMatrix) {

  glUseProgram(shaderProgram.program);

  // prepare modeling transform matrix
  glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), balon->position);
  modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
  modelMatrix = glm::scale(modelMatrix, glm::vec3(balon->size, balon->size, balon->size));

  // send matrices to the vertex & fragment shader
  setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

  setMaterialUniforms(
    balonGeometry->ambient,
    balonGeometry->diffuse,
    balonGeometry->specular,
    balonGeometry->shininess,
    balonGeometry->texture
  );

  // draw geometry
  glBindVertexArray(balonGeometry->vertexArrayObject);
  glDrawElements(GL_TRIANGLES, balonGeometry->numTriangles * 3, GL_UNSIGNED_INT, 0);

  glBindVertexArray(0);
  glUseProgram(0);

  return;
}

void drawDuck(DuckObject* duck, const glm::mat4 & viewMatrix, const glm::mat4 & projectionMatrix) {
  float angle = duck->rotationSpeed * (duck->currentTime-duck->startTime); // angle in radians

  glUseProgram(shaderProgram.program);

  glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), duck->position); //alignObject(duck->position, duck->direction, glm::vec3(0.0f, 0.0f, 1.0f))
  modelMatrix = glm::scale(modelMatrix, glm::vec3(duck->size));
  //modelMatrix = glm::rotate(modelMatrix, glm::radians(-90.0f), glm::vec3(1, 0, 0));
  

  // send matrices to the vertex & fragment shader
  setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

  setMaterialUniforms(
    duckGeometry->ambient,
    duckGeometry->diffuse,
    duckGeometry->specular,
    duckGeometry->shininess,
    duckGeometry->texture
  );

  // draw geometry
  glBindVertexArray(duckGeometry->vertexArrayObject);
  glDrawElements(GL_TRIANGLES, duckGeometry->numTriangles * 3, GL_UNSIGNED_INT, 0);

  glBindVertexArray(0);
  glUseProgram(0);

  return;
}

void drawExtraDuck(DuckObject* duck, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {
    float angle = duck->rotationSpeed * (duck->currentTime - duck->startTime); // angle in radians

    glUseProgram(shaderProgram.program);

    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), duck->position); //alignObject(duck->position, duck->direction, glm::vec3(0.0f, 0.0f, 1.0f))
    modelMatrix = alignObject(duck->position, duck->direction, glm::vec3(0.0f, 0.0f, 1.0f));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(duck->size));
    modelMatrix = glm::rotate(modelMatrix, glm::radians(-90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::rotate(modelMatrix, glm::radians(-90.0f), glm::vec3(0, 0, 1));


    // send matrices to the vertex & fragment shader
    setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniforms(
        duckGeometry->ambient,
        duckGeometry->diffuse,
        duckGeometry->specular,
        duckGeometry->shininess,
        duckGeometry->texture
    );

    // draw geometry
    glBindVertexArray(duckGeometry->vertexArrayObject);
    glDrawElements(GL_TRIANGLES, duckGeometry->numTriangles * 3, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    return;
}

void drawPyramid(Object* pyramid, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {

    glUseProgram(shaderProgram.program);

    // prepare modeling transform matrix
    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), pyramid->position);
    modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(pyramid->size, pyramid->size, pyramid->size));

    // send matrices to the vertex & fragment shader
    setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniforms(
        pyramidGeometry->ambient,
        pyramidGeometry->diffuse,
        pyramidGeometry->specular,
        pyramidGeometry->shininess,
        pyramidGeometry->texture
    );

    // draw geometry
    glBindVertexArray(pyramidGeometry->vertexArrayObject);
    glDrawElements(GL_TRIANGLES, pyramidGeometry->numTriangles * 3, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    return;
}

void drawPyramid2(Object* pyramid, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {

    glUseProgram(shaderProgram.program);

    // prepare modeling transform matrix
    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), pyramid->position);
    modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(pyramid->size, pyramid->size, pyramid->size));

    // send matrices to the vertex & fragment shader
    setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniforms(
        pyramidGeometry2->ambient,
        pyramidGeometry2->diffuse,
        pyramidGeometry2->specular,
        pyramidGeometry2->shininess,
        pyramidGeometry2->texture
    );

    // draw geometry
    glBindVertexArray(pyramidGeometry2->vertexArrayObject);
    glDrawElements(GL_TRIANGLES, pyramidGeometry2->numTriangles * 3, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    return;
}

void drawPyramid3(Object* pyramid, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {

    glUseProgram(shaderProgram.program);

    // prepare modeling transform matrix
    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), pyramid->position);
    modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(pyramid->size, pyramid->size, pyramid->size));

    // send matrices to the vertex & fragment shader
    setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniforms(
        pyramidGeometry3->ambient,
        pyramidGeometry3->diffuse,
        pyramidGeometry3->specular,
        pyramidGeometry3->shininess,
        pyramidGeometry3->texture
    );

    // draw geometry
    glBindVertexArray(pyramidGeometry3->vertexArrayObject);
    glDrawElements(GL_TRIANGLES, pyramidGeometry3->numTriangles * 3, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    return;
}

void drawFlag(Object* flag, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix, int geo_index)
{
    glUseProgram(shaderProgram.program);

    // prepare modeling transform matrix
    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), flag->position);
    modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(flag->size, flag->size, flag->size));

    // send matrices to the vertex & fragment shader
    setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniforms(
        flagGeometryList.at(geo_index)->ambient,
        flagGeometryList.at(geo_index)->diffuse,
        flagGeometryList.at(geo_index)->specular,
        flagGeometryList.at(geo_index)->shininess,
        flagGeometryList.at(geo_index)->texture
    );

    // draw geometry
    glBindVertexArray(flagGeometryList.at(geo_index)->vertexArrayObject);
    glDrawElements(GL_TRIANGLES, flagGeometryList.at(geo_index)->numTriangles * 3, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    return;
}

void drawAloe(Object* aloe, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {

    glUseProgram(shaderProgram.program);

    // prepare modeling transform matrix
    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), aloe->position);
    modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(aloe->size, aloe->size, aloe->size));

    // send matrices to the vertex & fragment shader
    setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniforms(
        aloeGeometry->ambient,
        aloeGeometry->diffuse,
        aloeGeometry->specular,
        aloeGeometry->shininess,
        aloeGeometry->texture
    );

    // draw geometry
    glBindVertexArray(aloeGeometry->vertexArrayObject);
    glDrawElements(GL_TRIANGLES, aloeGeometry->numTriangles * 3, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    return;
}

void drawSand(SandObject* sand, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix)
{
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);

    glUseProgram(sandShaderProgram.program);

      // just take 3x3 rotation part of the view transform
    glm::mat4 billboardRotationMatrix = glm::mat4(
        viewMatrix[0],
        viewMatrix[1],
        viewMatrix[2],
        glm::vec4(0.0f, 0.0f, 0.0f, 1.0f)
    );
    // inverse view rotation
    billboardRotationMatrix = glm::transpose(billboardRotationMatrix);

    glm::mat4 matrix = glm::mat4(1.0f);
    matrix = glm::translate(matrix, sand->position);
    matrix = glm::scale(matrix, glm::vec3(sand->size));

    matrix = matrix * billboardRotationMatrix; // make billboard to face the camera
    glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * matrix;

    glUniformMatrix4fv(sandShaderProgram.VmatrixLocation, 1, GL_FALSE, glm::value_ptr(viewMatrix));   // view

    glUniformMatrix4fv(sandShaderProgram.PVMmatrixLocation, 1, GL_FALSE, glm::value_ptr(PVMmatrix));  // model-view-projection
    glUniform1f(sandShaderProgram.timeLocation, sand->currentTime - sand->startTime);
    glUniform1i(sandShaderProgram.texSamplerLocation, 0);
    glUniform1f(sandShaderProgram.frameDurationLocation, sand->frameDuration);

    glBindVertexArray(sandGeometry->vertexArrayObject);
    glBindTexture(GL_TEXTURE_2D, sandGeometry->texture);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, sandGeometry->numTriangles);

    glBindVertexArray(0);
    glUseProgram(0);

    glDisable(GL_BLEND);

    return;
}

void drawFeather(SandObject* feather, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix)
{
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);

    glUseProgram(featherShaderProgram.program);

    // version 1: inversion of the rotation part of the view matrix

      // just take 3x3 rotation part of the view transform
    glm::mat4 billboardRotationMatrix = glm::mat4(
        viewMatrix[0],
        viewMatrix[1],
        viewMatrix[2],
        glm::vec4(0.0f, 0.0f, 0.0f, 1.0f)
    );
    // inverse view rotation
    billboardRotationMatrix = glm::transpose(billboardRotationMatrix);

    glm::mat4 matrix = glm::mat4(1.0f);
    matrix = glm::translate(matrix, feather->position);
    matrix = glm::scale(matrix, glm::vec3(feather->size));

    matrix = matrix * billboardRotationMatrix; // make billboard to face the camera
    glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * matrix;

    glUniformMatrix4fv(featherShaderProgram.VmatrixLocation, 1, GL_FALSE, glm::value_ptr(viewMatrix));   // view

    glUniformMatrix4fv(featherShaderProgram.PVMmatrixLocation, 1, GL_FALSE, glm::value_ptr(PVMmatrix));  // model-view-projection
    glUniform1f(featherShaderProgram.timeLocation, feather->currentTime - feather->startTime);
    glUniform1i(featherShaderProgram.texSamplerLocation, 0);
    glUniform1f(featherShaderProgram.frameDurationLocation, feather->frameDuration);

    glBindVertexArray(featherGeometry->vertexArrayObject);
    glBindTexture(GL_TEXTURE_2D, featherGeometry->texture);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, featherGeometry->numTriangles);

    glBindVertexArray(0);
    glUseProgram(0);

    glDisable(GL_BLEND);

    return;
}

void drawAloeA(SandObject* aloe, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix)
{
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);

    glUseProgram(aloeAShaderProgram.program);

    // version 1: inversion of the rotation part of the view matrix

      // just take 3x3 rotation part of the view transform
    glm::mat4 billboardRotationMatrix = glm::mat4(
        viewMatrix[0],
        viewMatrix[1],
        viewMatrix[2],
        glm::vec4(0.0f, 0.0f, 0.0f, 1.0f)
    );
    // inverse view rotation
    billboardRotationMatrix = glm::transpose(billboardRotationMatrix);

    glm::mat4 matrix = glm::mat4(1.0f);
    matrix = glm::translate(matrix, aloe->position);
    matrix = glm::scale(matrix, glm::vec3(aloe->size));

    matrix = matrix * billboardRotationMatrix; // make billboard to face the camera
    glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * matrix;

    glUniformMatrix4fv(aloeAShaderProgram.VmatrixLocation, 1, GL_FALSE, glm::value_ptr(viewMatrix));   // view

    glUniformMatrix4fv(aloeAShaderProgram.PVMmatrixLocation, 1, GL_FALSE, glm::value_ptr(PVMmatrix));  // model-view-projection
    glUniform1f(aloeAShaderProgram.timeLocation, aloe->currentTime - aloe->startTime);
    glUniform1i(aloeAShaderProgram.texSamplerLocation, 0);
    glUniform1f(aloeAShaderProgram.frameDurationLocation, aloe->frameDuration);

    glBindVertexArray(aloeAGeometry->vertexArrayObject);
    glBindTexture(GL_TEXTURE_2D, aloeAGeometry->texture);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, aloeAGeometry->numTriangles);

    glBindVertexArray(0);
    glUseProgram(0);

    glDisable(GL_BLEND);

    return;
}

void drawSkybox(const glm::mat4 & viewMatrix, const glm::mat4 & projectionMatrix) {

  glUseProgram(skyboxFarPlaneShaderProgram.program);

  // compose transformations
  glm::mat4 matrix = projectionMatrix * viewMatrix;

  // create view rotation matrix by using view matrix with cleared translation
  glm::mat4 viewRotation = viewMatrix;
  viewRotation[3] = glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);

  // vertex shader will translate screen space coordinates (NDC) using inverse PV matrix
  glm::mat4 inversePVmatrix = glm::inverse(projectionMatrix * viewRotation);

  glUniformMatrix4fv(skyboxFarPlaneShaderProgram.inversePVmatrixLocation, 1, GL_FALSE, glm::value_ptr(inversePVmatrix));
  glUniform1i(skyboxFarPlaneShaderProgram.skyboxSamplerLocation, 0);

  // draw "skybox" rendering 2 triangles covering the far plane
  glBindVertexArray(skyboxGeometry->vertexArrayObject);
  glBindTexture(GL_TEXTURE_CUBE_MAP, skyboxGeometry->texture);
  glDrawArrays(GL_TRIANGLE_STRIP, 0, skyboxGeometry->numTriangles+2);

  glBindVertexArray(0);
  glUseProgram(0);
}

void drawRock(Object* rock, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {

    glUseProgram(shaderProgram.program);

    // prepare modeling transform matrix
    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), rock->position);
    modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(rock->size, rock->size, rock->size));

    // send matrices to the vertex & fragment shader
    setTransformUniforms(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniforms(
        rockGeometry->ambient,
        rockGeometry->diffuse,
        rockGeometry->specular,
        rockGeometry->shininess,
        rockGeometry->texture
    );

    // draw geometry
    glBindVertexArray(rockGeometry->vertexArrayObject);
    glDrawElements(GL_TRIANGLES, rockGeometry->numTriangles * 3, GL_UNSIGNED_INT, 0);

    glBindVertexArray(0);
    glUseProgram(0);

    return;
}

void drawTexture(Object* texture, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix) {
   
    glEnable(GL_BLEND);
    CHECK_GL_ERROR();
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    CHECK_GL_ERROR();

    //glDisable(GL_DEPTH_TEST);
    glEnable(GL_DEPTH_TEST);
    CHECK_GL_ERROR();

    glUseProgram(texture1ShaderProgram.program);
    CHECK_GL_ERROR(); 

    glm::mat4 matrix = glm::translate(glm::mat4(1.0f), texture->position);
    matrix = glm::scale(matrix, glm::vec3(texture->size));

    glm::mat4 PVMmatrix = projectionMatrix * viewMatrix * matrix;

    GLuint location = glGetUniformLocation(texture1ShaderProgram.program, "isFog");
    glUniform1f(location, fog);

    glUniformMatrix4fv(texture1ShaderProgram.PVMmatrixLocation, 1, GL_FALSE, glm::value_ptr(PVMmatrix));        // model-view-projection
    glUniform1f(texture1ShaderProgram.timeLocation, texture->currentTime - texture->startTime);
    glUniform1i(texture1ShaderProgram.texSamplerLocation, 0);

    glBindTexture(GL_TEXTURE_2D, textureGeometry->texture);
    CHECK_GL_ERROR();
    glBindVertexArray(textureGeometry->vertexArrayObject);
    CHECK_GL_ERROR();

    glDrawArrays(GL_TRIANGLE_STRIP, 0, textureGeometry->numTriangles);
    CHECK_GL_ERROR();

    glBindVertexArray(0);
    glUseProgram(0);

    glEnable(GL_DEPTH_TEST);
    glDisable(GL_BLEND);

    return;
}

void drawCarpet(Object* carpet, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix)
{
    glUseProgram(carpetShaderProgram.program);

    // prepare modeling transform matrix
    glm::mat4 modelMatrix = glm::translate(glm::mat4(1.0f), carpet->position);
    //modelMatrix = glm::rotate(modelMatrix, glm::radians(90.0f), glm::vec3(1, 0, 0));
    modelMatrix = glm::scale(modelMatrix, glm::vec3(carpet->size, carpet->size, carpet->size));

    // send matrices to the vertex & fragment shader
    setTransformUniformsCarpet(modelMatrix, viewMatrix, projectionMatrix);

    setMaterialUniformsCarpet(
        carpetGeometry->ambient,
        carpetGeometry->diffuse,
        carpetGeometry->specular,
        carpetGeometry->shininess,
        carpetGeometry->texture
    );

    glUniform1f(carpetShaderProgram.timeLocation, carpet->currentTime - carpet->startTime);

    // draw geometry
    glBindVertexArray(carpetGeometry->vertexArrayObject);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, carpetGeometry->numTriangles);
    CHECK_GL_ERROR();
    glBindVertexArray(0);
    glUseProgram(0);
    CHECK_GL_ERROR();
    //std::cout << "I am in draw carpet" << std::endl;
    return;
}

void cleanupShaderPrograms(void) {

  pgr::deleteProgramAndShaders(shaderProgram.program);

  pgr::deleteProgramAndShaders(sandShaderProgram.program);
  pgr::deleteProgramAndShaders(featherShaderProgram.program);
  pgr::deleteProgramAndShaders(aloeAShaderProgram.program);
  pgr::deleteProgramAndShaders(skyboxFarPlaneShaderProgram.program);
  pgr::deleteProgramAndShaders(texture1ShaderProgram.program);
}

void initializeShaderPrograms(void) {

  std::vector<GLuint> shaderList;

  if(useLighting == true) {
    // load and compile shader for lighting (lights & materials)

    // push vertex shader and fragment shader
    shaderList.push_back(pgr::createShaderFromFile(GL_VERTEX_SHADER, "lightingPerVertex.vert"));
    shaderList.push_back(pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "lightingPerVertex.frag"));

    // create the shader program with two shaders
    shaderProgram.program = pgr::createProgram(shaderList);

    // get vertex attributes locations, if the shader does not have this uniform -> return -1
    shaderProgram.posLocation      = glGetAttribLocation(shaderProgram.program, "position");
    shaderProgram.normalLocation   = glGetAttribLocation(shaderProgram.program, "normal");
    shaderProgram.texCoordLocation = glGetAttribLocation(shaderProgram.program, "texCoord");
    // get uniforms locations
    shaderProgram.PVMmatrixLocation    = glGetUniformLocation(shaderProgram.program, "PVMmatrix");
    shaderProgram.VmatrixLocation      = glGetUniformLocation(shaderProgram.program, "Vmatrix");
    shaderProgram.MmatrixLocation      = glGetUniformLocation(shaderProgram.program, "Mmatrix");
    shaderProgram.normalMatrixLocation = glGetUniformLocation(shaderProgram.program, "normalMatrix");
    shaderProgram.timeLocation         = glGetUniformLocation(shaderProgram.program, "time");
    // material
    shaderProgram.ambientLocation      = glGetUniformLocation(shaderProgram.program, "material.ambient");
    shaderProgram.diffuseLocation      = glGetUniformLocation(shaderProgram.program, "material.diffuse");
    shaderProgram.specularLocation     = glGetUniformLocation(shaderProgram.program, "material.specular");
    shaderProgram.shininessLocation    = glGetUniformLocation(shaderProgram.program, "material.shininess");
    // texture
    shaderProgram.texSamplerLocation   = glGetUniformLocation(shaderProgram.program, "texSampler");
    shaderProgram.useTextureLocation   = glGetUniformLocation(shaderProgram.program, "material.useTexture");
    // reflector
    shaderProgram.reflectorPositionLocation  = glGetUniformLocation(shaderProgram.program, "reflectorPosition");
    shaderProgram.reflectorDirectionLocation = glGetUniformLocation(shaderProgram.program, "reflectorDirection");

    //shaderProgram.fog = glGetUniformLocation(shaderProgram.program, "isFog");
  }
  else {
    // load and compile simple shader (colors only, no lights at all)

    // push vertex shader and fragment shader
    shaderList.push_back(pgr::createShaderFromSource(GL_VERTEX_SHADER, colorVertexShaderSrc));
    shaderList.push_back(pgr::createShaderFromSource(GL_FRAGMENT_SHADER, colorFragmentShaderSrc));

    // create the program with two shaders (fragment and vertex)
    shaderProgram.program = pgr::createProgram(shaderList);
    // get position and color attributes locations
    shaderProgram.posLocation   = glGetAttribLocation(shaderProgram.program, "position");
    shaderProgram.colorLocation = glGetAttribLocation(shaderProgram.program, "color");
    // get uniforms locations
    shaderProgram.PVMmatrixLocation = glGetUniformLocation(shaderProgram.program, "PVMmatrix");

  }


  // load and compile shader for sand (dynamic texture)

  shaderList.clear();
  
  // push vertex shader and fragment shader
  shaderList.push_back(pgr::createShaderFromFile(GL_VERTEX_SHADER, "sand.vert"));
  shaderList.push_back(pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "sand.frag"));

  // create the program with two shaders
  sandShaderProgram.program = pgr::createProgram(shaderList);

  // get position and texture coordinates attributes locations
  sandShaderProgram.posLocation      = glGetAttribLocation(sandShaderProgram.program, "position");
  sandShaderProgram.texCoordLocation = glGetAttribLocation(sandShaderProgram.program, "texCoord");
  // get uniforms locations
  sandShaderProgram.PVMmatrixLocation     = glGetUniformLocation(sandShaderProgram.program, "PVMmatrix");
  sandShaderProgram.VmatrixLocation       = glGetUniformLocation(sandShaderProgram.program, "Vmatrix");
  sandShaderProgram.timeLocation          = glGetUniformLocation(sandShaderProgram.program, "time");
  sandShaderProgram.texSamplerLocation    = glGetUniformLocation(sandShaderProgram.program, "texSampler");
  sandShaderProgram.frameDurationLocation = glGetUniformLocation(sandShaderProgram.program, "frameDuration");

  // load and compile shader for feather

  shaderList.clear();

  shaderList.push_back(pgr::createShaderFromFile(GL_VERTEX_SHADER, "feather.vert"));
  shaderList.push_back(pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "feather.frag"));

  // create the program with two shaders
  featherShaderProgram.program = pgr::createProgram(shaderList);

  // get position and texture coordinates attributes locations
  featherShaderProgram.posLocation = glGetAttribLocation(featherShaderProgram.program, "position");
  featherShaderProgram.texCoordLocation = glGetAttribLocation(featherShaderProgram.program, "texCoord");
  // get uniforms locations
  featherShaderProgram.PVMmatrixLocation = glGetUniformLocation(featherShaderProgram.program, "PVMmatrix");
  featherShaderProgram.VmatrixLocation = glGetUniformLocation(featherShaderProgram.program, "Vmatrix");
  featherShaderProgram.timeLocation = glGetUniformLocation(featherShaderProgram.program, "time");
  featherShaderProgram.texSamplerLocation = glGetUniformLocation(featherShaderProgram.program, "texSampler");
  featherShaderProgram.frameDurationLocation = glGetUniformLocation(featherShaderProgram.program, "frameDuration");

  shaderList.clear();

  shaderList.push_back(pgr::createShaderFromFile(GL_VERTEX_SHADER, "aloe.vert"));
  shaderList.push_back(pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "aloe.frag"));

  // create the program with two shaders
  aloeAShaderProgram.program = pgr::createProgram(shaderList);

  // get position and texture coordinates attributes locations
  aloeAShaderProgram.posLocation = glGetAttribLocation(aloeAShaderProgram.program, "position");
  aloeAShaderProgram.texCoordLocation = glGetAttribLocation(aloeAShaderProgram.program, "texCoord");
  // get uniforms locations
  aloeAShaderProgram.PVMmatrixLocation = glGetUniformLocation(aloeAShaderProgram.program, "PVMmatrix");
  aloeAShaderProgram.VmatrixLocation = glGetUniformLocation(aloeAShaderProgram.program, "Vmatrix");
  aloeAShaderProgram.timeLocation = glGetUniformLocation(aloeAShaderProgram.program, "time");
  aloeAShaderProgram.texSamplerLocation = glGetUniformLocation(aloeAShaderProgram.program, "texSampler");
  aloeAShaderProgram.frameDurationLocation = glGetUniformLocation(aloeAShaderProgram.program, "frameDuration");

  shaderList.clear();

  // push vertex shader and fragment shader
  shaderList.push_back(pgr::createShaderFromSource(GL_VERTEX_SHADER, skyboxFarPlaneVertexShaderSrc));
  shaderList.push_back(pgr::createShaderFromSource(GL_FRAGMENT_SHADER, skyboxFarPlaneFragmentShaderSrc));

  // create the program with two shaders
  skyboxFarPlaneShaderProgram.program = pgr::createProgram(shaderList);

  // handles to vertex attributes locations
  skyboxFarPlaneShaderProgram.screenCoordLocation = glGetAttribLocation(skyboxFarPlaneShaderProgram.program, "screenCoord");
  // get uniforms locations
  skyboxFarPlaneShaderProgram.skyboxSamplerLocation   = glGetUniformLocation(skyboxFarPlaneShaderProgram.program, "skyboxSampler");
  skyboxFarPlaneShaderProgram.inversePVmatrixLocation = glGetUniformLocation(skyboxFarPlaneShaderProgram.program, "inversePVmatrix");

  shaderList.clear();

  shaderList.push_back(pgr::createShaderFromFile(GL_VERTEX_SHADER, "texture1.vert"));
  shaderList.push_back(pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "texture1.frag"));
  CHECK_GL_ERROR();

  // Create the program with two shaders
  texture1ShaderProgram.program = pgr::createProgram(shaderList);
  CHECK_GL_ERROR();

  // get position and color attributes locations
  texture1ShaderProgram.posLocation = glGetAttribLocation(texture1ShaderProgram.program, "position");
  texture1ShaderProgram.texCoordLocation = glGetAttribLocation(texture1ShaderProgram.program, "texCoord");
  // get uniforms locations
  texture1ShaderProgram.PVMmatrixLocation = glGetUniformLocation(texture1ShaderProgram.program, "PVMmatrix");
  texture1ShaderProgram.timeLocation = glGetUniformLocation(texture1ShaderProgram.program, "time");
  texture1ShaderProgram.texSamplerLocation = glGetUniformLocation(texture1ShaderProgram.program, "texSampler");

  shaderList.clear();
  //------------------------------------------------------------------------------------------------------------------------
  if (useLighting == true) {
      // load and compile shader for lighting (lights & materials)

      // push vertex shader and fragment shader
      //shaderList.push_back(pgr::createShaderFromFile(GL_VERTEX_SHADER, "lightingPerVertex.vert"));
      //shaderList.push_back(pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "lightingPerVertex.frag"));
      shaderList.push_back(pgr::createShaderFromFile(GL_VERTEX_SHADER, "carpet.vert"));
      shaderList.push_back(pgr::createShaderFromFile(GL_FRAGMENT_SHADER, "carpet.frag"));

      // create the shader program with two shaders
      carpetShaderProgram.program = pgr::createProgram(shaderList);

      // get vertex attributes locations, if the shader does not have this uniform -> return -1
      carpetShaderProgram.posLocation = glGetAttribLocation(carpetShaderProgram.program, "position");
      carpetShaderProgram.normalLocation = glGetAttribLocation(carpetShaderProgram.program, "normal");
      carpetShaderProgram.texCoordLocation = glGetAttribLocation(carpetShaderProgram.program, "texCoord");
      // get uniforms locations
      carpetShaderProgram.PVMmatrixLocation = glGetUniformLocation(carpetShaderProgram.program, "PVMmatrix");
      carpetShaderProgram.VmatrixLocation = glGetUniformLocation(carpetShaderProgram.program, "Vmatrix");
      carpetShaderProgram.MmatrixLocation = glGetUniformLocation(carpetShaderProgram.program, "Mmatrix");
      carpetShaderProgram.normalMatrixLocation = glGetUniformLocation(carpetShaderProgram.program, "normalMatrix");
      carpetShaderProgram.timeLocation = glGetUniformLocation(carpetShaderProgram.program, "time");
      // material
      carpetShaderProgram.ambientLocation = glGetUniformLocation(carpetShaderProgram.program, "material.ambient");
      carpetShaderProgram.diffuseLocation = glGetUniformLocation(carpetShaderProgram.program, "material.diffuse");
      carpetShaderProgram.specularLocation = glGetUniformLocation(carpetShaderProgram.program, "material.specular");
      carpetShaderProgram.shininessLocation = glGetUniformLocation(carpetShaderProgram.program, "material.shininess");
      // texture
      carpetShaderProgram.texSamplerLocation = glGetUniformLocation(carpetShaderProgram.program, "texSampler");
      carpetShaderProgram.useTextureLocation = glGetUniformLocation(carpetShaderProgram.program, "material.useTexture");
      // reflector
      carpetShaderProgram.reflectorPositionLocation = glGetUniformLocation(carpetShaderProgram.program, "reflectorPosition");
      carpetShaderProgram.reflectorDirectionLocation = glGetUniformLocation(carpetShaderProgram.program, "reflectorDirection");

      //std::cout << "I am in shader carpet" << std::endl;
      //shaderProgram.fog = glGetUniformLocation(shaderProgram.program, "isFog");
  }
  else {
      // load and compile simple shader (colors only, no lights at all)

      // push vertex shader and fragment shader
      shaderList.push_back(pgr::createShaderFromSource(GL_VERTEX_SHADER, colorVertexShaderSrc));
      shaderList.push_back(pgr::createShaderFromSource(GL_FRAGMENT_SHADER, colorFragmentShaderSrc));

      // create the program with two shaders (fragment and vertex)
      carpetShaderProgram.program = pgr::createProgram(shaderList);
      // get position and color attributes locations
      carpetShaderProgram.posLocation = glGetAttribLocation(carpetShaderProgram.program, "position");
      carpetShaderProgram.colorLocation = glGetAttribLocation(carpetShaderProgram.program, "color");
      // get uniforms locations
      carpetShaderProgram.PVMmatrixLocation = glGetUniformLocation(carpetShaderProgram.program, "PVMmatrix");

  }

}

/** Load mesh using assimp library
 *  Vertex, normals and texture coordinates data are stored without interleaving |VVVVV...|NNNNN...|tttt
 * \param fileName [in] file to open/load
 * \param shader [in] vao will connect loaded data to shader
 * \param geometry
 */
bool loadSingleMesh(const std::string &fileName, SCommonShaderProgram& shader, MeshGeometry** geometry) {
  Assimp::Importer importer;

  // Unitize object in size (scale the model to fit into (-1..1)^3)
  importer.SetPropertyInteger(AI_CONFIG_PP_PTV_NORMALIZE, 1);

  // Load asset from the file - you can play with various processing steps
  const aiScene * scn = importer.ReadFile(fileName.c_str(), 0
      | aiProcess_Triangulate             // Triangulate polygons (if any).
      | aiProcess_PreTransformVertices    // Transforms scene hierarchy into one root with geometry-leafs only. For more see Doc.
      | aiProcess_GenSmoothNormals        // Calculate normals per vertex.
      | aiProcess_JoinIdenticalVertices);

  // abort if the loader fails
  if(scn == NULL) {
    std::cerr << "assimp error: " << importer.GetErrorString() << std::endl;
    *geometry = NULL;
    return false;
  }

  // some formats store whole scene (multiple meshes and materials, lights, cameras, ...) in one file, we cannot handle that in our simplified example
  if(scn->mNumMeshes != 1) {
    std::cerr << "this simplified loader can only process files with only one mesh" << std::endl;
    *geometry = NULL;
    return false;
  }

  // in this phase we know we have one mesh in our loaded scene, we can directly copy its data to OpenGL ...
  const aiMesh * mesh = scn->mMeshes[0];

  *geometry = new MeshGeometry;

  // vertex buffer object, store all vertex positions and normals
  glGenBuffers(1, &((*geometry)->vertexBufferObject));
  glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
  glBufferData(GL_ARRAY_BUFFER, 8*sizeof(float)*mesh->mNumVertices, 0, GL_STATIC_DRAW); // allocate memory for vertices, normals, and texture coordinates
  // first store all vertices
  glBufferSubData(GL_ARRAY_BUFFER, 0, 3*sizeof(float)*mesh->mNumVertices, mesh->mVertices);
  // then store all normals
  glBufferSubData(GL_ARRAY_BUFFER, 3*sizeof(float)*mesh->mNumVertices, 3*sizeof(float)*mesh->mNumVertices, mesh->mNormals);
  
  // just texture 0 for now
  float *textureCoords = new float[2 * mesh->mNumVertices];  // 2 floats per vertex
  float *currentTextureCoord = textureCoords;

  // copy texture coordinates
  aiVector3D vect;
    
  if(mesh->HasTextureCoords(0) ) {
    // we use 2D textures with 2 coordinates and ignore the third coordinate
    for(unsigned int idx=0; idx<mesh->mNumVertices; idx++) {
      vect = (mesh->mTextureCoords[0])[idx];
      *currentTextureCoord++ = vect.x;
      *currentTextureCoord++ = vect.y;
    }
  }
    
  // finally store all texture coordinates
  glBufferSubData(GL_ARRAY_BUFFER, 6*sizeof(float)*mesh->mNumVertices, 2*sizeof(float)*mesh->mNumVertices, textureCoords);

  // copy all mesh faces into one big array (assimp supports faces with ordinary number of vertices, we use only 3 -> triangles)
  unsigned int *indices = new unsigned int[mesh->mNumFaces * 3];
  for(unsigned int f = 0; f < mesh->mNumFaces; ++f) {
    indices[f*3 + 0] = mesh->mFaces[f].mIndices[0];
    indices[f*3 + 1] = mesh->mFaces[f].mIndices[1];
    indices[f*3 + 2] = mesh->mFaces[f].mIndices[2];
  }

  // copy our temporary index array to OpenGL and free the array
  glGenBuffers(1, &((*geometry)->elementBufferObject));
  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, (*geometry)->elementBufferObject);
  glBufferData(GL_ELEMENT_ARRAY_BUFFER, 3 * sizeof(unsigned) * mesh->mNumFaces, indices, GL_STATIC_DRAW);

  delete [] indices;

  // copy the material info to MeshGeometry structure
  const aiMaterial *mat  = scn->mMaterials[mesh->mMaterialIndex];
  aiColor4D color;
  aiString name;
  aiReturn retValue = AI_SUCCESS;

  // Get returns: aiReturn_SUCCESS 0 | aiReturn_FAILURE -1 | aiReturn_OUTOFMEMORY -3
  mat->Get(AI_MATKEY_NAME, name); // may be "" after the input mesh processing. Must be aiString type!

  if((retValue = aiGetMaterialColor(mat, AI_MATKEY_COLOR_DIFFUSE, &color)) != AI_SUCCESS)
    color = aiColor4D(0.0f, 0.0f, 0.0f, 0.0f);

  (*geometry)->diffuse = glm::vec3(color.r, color.g, color.b);

  if ((retValue = aiGetMaterialColor(mat, AI_MATKEY_COLOR_AMBIENT, &color)) != AI_SUCCESS)
    color = aiColor4D(0.0f, 0.0f, 0.0f, 0.0f);
  (*geometry)->ambient = glm::vec3(color.r, color.g, color.b);

  if ((retValue = aiGetMaterialColor(mat, AI_MATKEY_COLOR_SPECULAR, &color)) != AI_SUCCESS)
    color = aiColor4D(0.0f, 0.0f, 0.0f, 0.0f);
  (*geometry)->specular = glm::vec3(color.r, color.g, color.b);

  ai_real shininess, strength;
  unsigned int max;	// changed: to unsigned

  max = 1;	
  if ((retValue = aiGetMaterialFloatArray(mat, AI_MATKEY_SHININESS, &shininess, &max)) != AI_SUCCESS)
    shininess = 1.0f;
  max = 1;
  if((retValue = aiGetMaterialFloatArray(mat, AI_MATKEY_SHININESS_STRENGTH, &strength, &max)) != AI_SUCCESS)
    strength = 1.0f;
  (*geometry)->shininess = shininess * strength;

  (*geometry)->texture = 0;

  // load texture image
  if (mat->GetTextureCount(aiTextureType_DIFFUSE) > 0) {
    // get texture name 
    aiString path; // filename

    aiReturn texFound = mat->GetTexture(aiTextureType_DIFFUSE, 0, &path);
    std::string textureName = path.data;

    size_t found = fileName.find_last_of("/\\");
    // insert correct texture file path 
    if(found != std::string::npos) { // not found
      //subMesh_p->textureName.insert(0, "/");
      textureName.insert(0, fileName.substr(0, found+1));
    }

    std::cout << "Loading texture file: " << textureName << std::endl;
    (*geometry)->texture = pgr::createTexture(textureName);
  }
  CHECK_GL_ERROR();

  glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
  glBindVertexArray((*geometry)->vertexArrayObject);

  glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, (*geometry)->elementBufferObject); // bind our element array buffer (indices) to vao
  glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);

  glEnableVertexAttribArray(shader.posLocation);
  glVertexAttribPointer(shader.posLocation, 3, GL_FLOAT, GL_FALSE, 0, 0);

  if(useLighting == true) {
    glEnableVertexAttribArray(shader.normalLocation);
    glVertexAttribPointer(shader.normalLocation, 3, GL_FLOAT, GL_FALSE, 0, (void*)(3 * sizeof(float) * mesh->mNumVertices));
  }
  else {
	  glDisableVertexAttribArray(shader.colorLocation);
	  // following line is problematic on AMD/ATI graphic cards
	  // -> if you see black screen (no objects at all) than try to set color manually in vertex shader to see at least something
    glVertexAttrib3f(shader.colorLocation, color.r, color.g, color.b);
  }

  glEnableVertexAttribArray(shader.texCoordLocation);
  glVertexAttribPointer(shader.texCoordLocation, 2, GL_FLOAT, GL_FALSE, 0, (void*)(6 * sizeof(float) * mesh->mNumVertices));
  CHECK_GL_ERROR();

  glBindVertexArray(0);

  (*geometry)->numTriangles = mesh->mNumFaces;

  return true;
}

//
// specific initialization of models geometry
//
void initSandGeometry(GLuint shader, MeshGeometry **geometry) {
    
  *geometry = new MeshGeometry;

  (*geometry)->texture = pgr::createTexture(SAND_TEXTURE_NAME);
  
  glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
  glBindVertexArray((*geometry)->vertexArrayObject);
  
  glGenBuffers(1, &((*geometry)->vertexBufferObject));\
  glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
  glBufferData(GL_ARRAY_BUFFER, sizeof(sandVertexData), sandVertexData, GL_STATIC_DRAW);
  
  glEnableVertexAttribArray(sandShaderProgram.posLocation);
  // vertices of triangles - start at the beginning of the array (interlaced array)
  glVertexAttribPointer(sandShaderProgram.posLocation, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), 0);
  //-------------------------------------------------- 
  glEnableVertexAttribArray(sandShaderProgram.texCoordLocation);
  CHECK_GL_ERROR();
  // texture coordinates are placed just after the position of each vertex (interlaced array)
  glVertexAttribPointer(sandShaderProgram.texCoordLocation, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
  //-----------------------------------------------------
  glBindVertexArray(0);

  (*geometry)->numTriangles = sandNumQuadVertices;
}

void initFeatherGeometry(GLuint shader, MeshGeometry** geometry) {

    *geometry = new MeshGeometry;

    (*geometry)->texture = pgr::createTexture(FEATHER_TEXTURE_NAME);

    glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
    glBindVertexArray((*geometry)->vertexArrayObject);

    glGenBuffers(1, &((*geometry)->vertexBufferObject)); \
        glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(featherVertices), featherVertices, GL_STATIC_DRAW);

    glEnableVertexAttribArray(featherShaderProgram.posLocation);
    // vertices of triangles - start at the beginning of the array (interlaced array)
    glVertexAttribPointer(featherShaderProgram.posLocation, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), 0);
    //-------------------------------------------------- 
    glEnableVertexAttribArray(featherShaderProgram.texCoordLocation);
    CHECK_GL_ERROR();
    // texture coordinates are placed just after the position of each vertex (interlaced array)
    glVertexAttribPointer(featherShaderProgram.texCoordLocation, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
    //-----------------------------------------------------
    glBindVertexArray(0);

    (*geometry)->numTriangles = featherNumQuadVertices;
}

void initAloeAGeometry(GLuint shader, MeshGeometry** geometry) {

    *geometry = new MeshGeometry;

    (*geometry)->texture = pgr::createTexture(ALOE_A_TEXTURE_NAME);

    glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
    glBindVertexArray((*geometry)->vertexArrayObject);

    glGenBuffers(1, &((*geometry)->vertexBufferObject)); \
        glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(aloeVertices), aloeVertices, GL_STATIC_DRAW);

    glEnableVertexAttribArray(aloeAShaderProgram.posLocation);
    // vertices of triangles - start at the beginning of the array (interlaced array)
    glVertexAttribPointer(aloeAShaderProgram.posLocation, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), 0);
    //-------------------------------------------------- 
    glEnableVertexAttribArray(aloeAShaderProgram.texCoordLocation);
    CHECK_GL_ERROR();
    // texture coordinates are placed just after the position of each vertex (interlaced array)
    glVertexAttribPointer(aloeAShaderProgram.texCoordLocation, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
    //-----------------------------------------------------
    glBindVertexArray(0);

    (*geometry)->numTriangles = aloeNumQuadVertices;
}

void initSkyboxGeometry(GLuint shader, MeshGeometry **geometry) {

  *geometry = new MeshGeometry;

  // 2D coordinates of 2 triangles covering the whole screen (NDC), draw using triangle strip
  static const float screenCoords[] = {
    -1.0f, -1.0f,
     1.0f, -1.0f,
    -1.0f,  1.0f,
     1.0f,  1.0f
  };

  glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
  glBindVertexArray((*geometry)->vertexArrayObject);

  // buffer for far plane rendering
  glGenBuffers(1, &((*geometry)->vertexBufferObject));\
  glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
  glBufferData(GL_ARRAY_BUFFER, sizeof(screenCoords), screenCoords, GL_STATIC_DRAW);

  //glUseProgram(farplaneShaderProgram);

  glEnableVertexAttribArray(skyboxFarPlaneShaderProgram.screenCoordLocation);
  glVertexAttribPointer(skyboxFarPlaneShaderProgram.screenCoordLocation, 2, GL_FLOAT, GL_FALSE, 0, 0);

  glBindVertexArray(0);
  glUseProgram(0);
  CHECK_GL_ERROR(); //CHYBA

  (*geometry)->numTriangles = 2;

  glActiveTexture(GL_TEXTURE0);

  glGenTextures(1, &((*geometry)->texture));
  glBindTexture(GL_TEXTURE_CUBE_MAP, (*geometry)->texture);

  const char * suffixes[] = { "posx", "negx", "posy", "negy", "posz", "negz" };
  GLuint targets[] = {
    GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Y, GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
    GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
  };

  for( int i = 0; i < 6; i++ ) {
    std::string texName = std::string(SKYBOX_CUBE_TEXTURE_FILE_PREFIX) + "_" + suffixes[i] + ".jpg";
    std::cout << "Loading cube map texture: " << texName << std::endl;
    if(!pgr::loadTexImage2D(texName, targets[i])) {
      pgr::dieWithError("Skybox cube map loading failed!");
    }
  }

  glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
  glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glTexParameterf(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
  glGenerateMipmap(GL_TEXTURE_CUBE_MAP);

  // unbind the texture (just in case someone will mess up with texture calls later)
  glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
  CHECK_GL_ERROR();
}

void initRock(SCommonShaderProgram& shader, MeshGeometry **geometry) {
    *geometry = new MeshGeometry();
    glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
    glBindVertexArray((*geometry)->vertexArrayObject);

    // Make vbo for vertexes
    glGenBuffers(1, &((*geometry)->vertexBufferObject));
    glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(rockVertices), rockVertices, GL_STATIC_DRAW);
    CHECK_GL_ERROR();

    //Bind locations
    glEnableVertexAttribArray(shader.posLocation);
    glVertexAttribPointer(shader.posLocation, 3, GL_FLOAT, GL_FALSE, rockNAttribsPerVertex * sizeof(float), 0);

    // color of vertex starts after the position (interlaced arrays)
    glEnableVertexAttribArray(shader.normalLocation);
    glVertexAttribPointer(shader.normalLocation, 3, GL_FLOAT, GL_FALSE, rockNAttribsPerVertex * sizeof(float), (void*)(3 * sizeof(float)));
    CHECK_GL_ERROR();
    glEnableVertexAttribArray(shader.texCoordLocation);
    glVertexAttribPointer(shader.texCoordLocation, 2, GL_FLOAT, GL_FALSE, rockNAttribsPerVertex * sizeof(float), (void*)(6 * sizeof(float)));
    CHECK_GL_ERROR();

    //ebo
    glGenBuffers(1, &((*geometry)->elementBufferObject));
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, (*geometry)->elementBufferObject);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(rockIndices), rockIndices, GL_STATIC_DRAW);
    CHECK_GL_ERROR();

    (*geometry)->ambient = glm::vec3(0.37890625f, 0.33984375f, 0.24609375f); // 97, 87, 63
    (*geometry)->diffuse = glm::vec3(0.37890625f, 0.33984375f, 0.24609375f); //0.37890625f, 0.33984375f, 0.24609375f
    (*geometry)->specular = glm::vec3(0.37890625f, 0.33984375f, 0.24609375f);
    (*geometry)->shininess = 10.0f;
    (*geometry)->texture = 0;

    (*geometry)->numTriangles = rockNTriangles;
    glBindVertexArray(0);
}

void initTextureGeometry(GLuint shader, MeshGeometry** geometry) {

    *geometry = new MeshGeometry;

    (*geometry)->texture = pgr::createTexture(MOVING_TEXTURE_NAME);
    //(*geometry)->ambient = glm::vec3(0.37890625f, 0.33984375f, 0.24609375f); // 97, 87, 63
    CHECK_GL_ERROR();
    glBindTexture(GL_TEXTURE_2D, (*geometry)->texture);
    CHECK_GL_ERROR();

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
    CHECK_GL_ERROR();

    glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
    glBindVertexArray((*geometry)->vertexArrayObject);
    CHECK_GL_ERROR();

    glGenBuffers(1, &((*geometry)->vertexBufferObject));
    glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(texture1VertexData), texture1VertexData, GL_STATIC_DRAW);
    CHECK_GL_ERROR();

    glEnableVertexAttribArray(texture1ShaderProgram.posLocation);
    glEnableVertexAttribArray(texture1ShaderProgram.texCoordLocation);
    // vertices of triangles - start at the beginning of the interlaced array
    glVertexAttribPointer(texture1ShaderProgram.posLocation, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), 0);
    // texture coordinates of each vertices are stored just after its position
    glVertexAttribPointer(texture1ShaderProgram.texCoordLocation, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
    CHECK_GL_ERROR();

    glBindVertexArray(0);

    (*geometry)->numTriangles = texture1NumQuadVertices;
}

void initCarpetGeometry(SCommonShaderProgram& shader, MeshGeometry** geometry)
{
    generateCarpetVertexData();

    *geometry = new MeshGeometry;

    glGenVertexArrays(1, &((*geometry)->vertexArrayObject));
    glBindVertexArray((*geometry)->vertexArrayObject);

    // Make vbo for vertexes
    glGenBuffers(1, &((*geometry)->vertexBufferObject));
    glBindBuffer(GL_ARRAY_BUFFER, (*geometry)->vertexBufferObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(carpetVertexData), carpetVertexData, GL_STATIC_DRAW);
    CHECK_GL_ERROR();

    //Bind locations
    glEnableVertexAttribArray(shader.posLocation);
    glVertexAttribPointer(shader.posLocation, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), 0);

    // color of vertex starts after the position (interlaced arrays)
    glEnableVertexAttribArray(shader.normalLocation);
    glVertexAttribPointer(shader.normalLocation, 3, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(3 * sizeof(float)));
    CHECK_GL_ERROR();
    glEnableVertexAttribArray(shader.texCoordLocation);
    glVertexAttribPointer(shader.texCoordLocation, 2, GL_FLOAT, GL_FALSE, 5 * sizeof(float), (void*)(6 * sizeof(float)));
    CHECK_GL_ERROR();

    (*geometry)->ambient = glm::vec3(0.8f, 0.4f, 0.0f); // 97, 87, 63
    (*geometry)->diffuse = glm::vec3(0.8f, 0.40f, 0.0f); //0.37890625f, 0.33984375f, 0.24609375f
    (*geometry)->specular = glm::vec3(0.8f, 0.40f, 0.0f);
    (*geometry)->shininess = 5.0f;
    (*geometry)->texture = pgr::createTexture(CARPET_TEXTURE_NAME);
    CHECK_GL_ERROR();
    (*geometry)->numTriangles = (CARPET_SLICE_COUNT + 1) * 2;;
    glBindVertexArray(0);
}

//
// end of specific initialization of models geometry
//

/* Initialize vertex buffers and vertex arrays for all objects. 
 */
void initializeModels() {

  // load duck model from external file
  if(loadSingleMesh(DUCK_MODEL_NAME, shaderProgram, &duckGeometry) != true) {
    std::cerr << "initializeModels(): Duck model loading failed." << std::endl;
  }
  CHECK_GL_ERROR();

  // load balon model from external file
  if(loadSingleMesh(BALON_MODEL_NAME, shaderProgram, &balonGeometry) != true) {
    std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  CHECK_GL_ERROR();

  // load pyrammids models from external file
  if (loadSingleMesh(PYRAMID_1_MODEL_NAME, shaderProgram, &pyramidGeometry) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  CHECK_GL_ERROR();

  if (loadSingleMesh(PYRAMID_2_MODEL_NAME, shaderProgram, &pyramidGeometry2) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  CHECK_GL_ERROR();

  if (loadSingleMesh(PYRAMID_3_MODEL_NAME, shaderProgram, &pyramidGeometry3) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  CHECK_GL_ERROR();

  // load flag model from external file
  if (loadSingleMesh(FLAG_MODEL_NAME_2, shaderProgram, &flagGeometry) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  flagGeometryList.push_back(flagGeometry);
  if (loadSingleMesh(FLAG_MODEL_NAME_3, shaderProgram, &flagGeometry) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  flagGeometryList.push_back(flagGeometry);
  if (loadSingleMesh(FLAG_MODEL_NAME_4, shaderProgram, &flagGeometry) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  flagGeometryList.push_back(flagGeometry);
  if (loadSingleMesh(FLAG_MODEL_NAME_5, shaderProgram, &flagGeometry) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  flagGeometryList.push_back(flagGeometry);
  if (loadSingleMesh(FLAG_MODEL_NAME_6, shaderProgram, &flagGeometry) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  flagGeometryList.push_back(flagGeometry);
  CHECK_GL_ERROR();

  //load aloe model
  if (loadSingleMesh(ALOE_MODEL_NAME, shaderProgram, &aloeGeometry) != true) {
      std::cerr << "initializeModels(): Space ship model loading failed." << std::endl;
  }
  CHECK_GL_ERROR();

  //load sand
  initSandGeometry(sandShaderProgram.program, &sandGeometry);
  CHECK_GL_ERROR();

  //load feather
  initFeatherGeometry(featherShaderProgram.program, &featherGeometry);
  CHECK_GL_ERROR();

  //load aloe animation
  initAloeAGeometry(aloeAShaderProgram.program, &aloeAGeometry);
  CHECK_GL_ERROR();

  //init rock
  initRock(shaderProgram, &rockGeometry);

  initTextureGeometry(texture1ShaderProgram.program, &textureGeometry);

  // fill MeshGeometry structure for skybox object
  initSkyboxGeometry(skyboxFarPlaneShaderProgram.program, &skyboxGeometry);

  //init carpet
  initCarpetGeometry(carpetShaderProgram, &carpetGeometry);
}

void cleanupGeometry(MeshGeometry *geometry) {

  glDeleteVertexArrays(1, &(geometry->vertexArrayObject));
  glDeleteBuffers(1, &(geometry->elementBufferObject));
  glDeleteBuffers(1, &(geometry->vertexBufferObject));

  if(geometry->texture != 0)
    glDeleteTextures(1, &(geometry->texture));
}

void cleanupModels() {

  cleanupGeometry(balonGeometry);
  cleanupGeometry(duckGeometry);

  cleanupGeometry(pyramidGeometry);
  cleanupGeometry(pyramidGeometry2);
  cleanupGeometry(pyramidGeometry3);

  cleanupGeometry(flagGeometry);
  cleanupGeometry(aloeGeometry);
  cleanupGeometry(sandGeometry);
  cleanupGeometry(featherGeometry);
  cleanupGeometry(rockGeometry);
  cleanupGeometry(textureGeometry);
  cleanupGeometry(aloeAGeometry);
  cleanupGeometry(carpetGeometry);

  cleanupGeometry(skyboxGeometry);
  
}
