//----------------------------------------------------------------------------------------
/**
 * \file    render_stuff.h
 * \author  Zuzana Slaninova
 * \using   asteroids
 * \date    2023
 * \brief   Rendering stuff - drawing functions for models, etc.
 */
//----------------------------------------------------------------------------------------

#ifndef __RENDER_STUFF_H
#define __RENDER_STUFF_H

#include "data.h"

// defines geometry of object in the scene
// geometry is shared among all instances of the same object type
typedef struct _MeshGeometry {
  GLuint        vertexBufferObject;   // identifier for the vertex buffer object
  GLuint        elementBufferObject;  // identifier for the element buffer object
  GLuint        vertexArrayObject;    // identifier for the vertex array object
  unsigned int  numTriangles;         // number of triangles in the mesh
  // material
  glm::vec3     ambient;
  glm::vec3     diffuse;
  glm::vec3     specular;
  float         shininess;
  GLuint        texture;

} MeshGeometry;

// parameters of individual objects in the scene (e.g. position, size, speed, etc.)
typedef struct _Object {
  glm::vec3 position;
  glm::vec3 direction;
  float     speed;
  float     size;

  bool destroyed;

  float startTime;
  float currentTime;

} Object;

typedef struct _BalonObject : public Object {

  float viewAngle; // in degrees

} BalonObject;

typedef struct _DuckObject : public Object {

  float rotationSpeed;

} DuckObject;


typedef struct _SandObject : public Object {

  int    textureFrames;
  float  frameDuration;

} SandObject;


typedef struct _commonShaderProgram {
  // identifier for the shader program
  GLuint program;          // = 0;
  // vertex attributes locations
  GLint posLocation;       // = -1;
  GLint colorLocation;     // = -1;
  GLint normalLocation;    // = -1;
  GLint texCoordLocation;  // = -1;
  // uniforms locations
  GLint PVMmatrixLocation;    // = -1;
  GLint VmatrixLocation;      // = -1;  view/camera matrix
  GLint MmatrixLocation;      // = -1;  modeling matrix
  GLint normalMatrixLocation; // = -1;  inverse transposed Mmatrix

  GLint timeLocation;         // = -1; elapsed time in seconds

  // material 
  GLint diffuseLocation;    // = -1;
  GLint ambientLocation;    // = -1;
  GLint specularLocation;   // = -1;
  GLint shininessLocation;  // = -1;
  // texture
  GLint useTextureLocation; // = -1; 
  GLint texSamplerLocation; // = -1;
  // reflector related uniforms
  GLint reflectorPositionLocation;  // = -1; 
  GLint reflectorDirectionLocation; // = -1;

  bool fog; //false
} SCommonShaderProgram;

//draw balon from object in view and projection matrix
void drawBalon(BalonObject* balon, const glm::mat4 & viewMatrix, const glm::mat4 & projectionMatrix);

//draw duck from object in view and projection matrix
void drawDuck(DuckObject* duck, const glm::mat4 & viewMatrix, const glm::mat4 & projectionMatrix);

//draw extraDuck from object in view and projection matrix
void drawExtraDuck(DuckObject* duck, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw sand billboard from object in view and projection matrix
void drawSand(SandObject* sand, const glm::mat4 & viewMatrix, const glm::mat4 & projectionMatrix);

//draw feather billboard from object in view and projection matrix
void drawFeather(SandObject* feather, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw aloe billboard from object in view and projection matrix
void drawAloeA(SandObject* aloe, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw skybox in view and projection matrix
void drawSkybox(const glm::mat4 & viewMatrix, const glm::mat4 & projectionMatrix);

//draw pyramid  type 1 from object in view and projection matrix
void drawPyramid(Object* pyramid, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw pyramid type 2 from object in view and projection matrix
void drawPyramid2(Object* pyramid, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw pyramid type 3 from object in view and projection matrix
void drawPyramid3(Object* pyramid, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw flag from object in view and projection matrix
void drawFlag(Object* flag, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix, int geo_index);

//draw aloe from object in view and projection matrix
void drawAloe(Object* aloe, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw rock from object in view and projection matrix
void drawRock(Object* rock, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw moving texture from object in view and projection matrix
void drawTexture(Object* texture, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

//draw carpet from object in view and projection matrix
void drawCarpet(Object* carpet, const glm::mat4& viewMatrix, const glm::mat4& projectionMatrix);

void initializeShaderPrograms();

void cleanupShaderPrograms();

void initializeModels();

void cleanupModels();

#endif // __RENDER_STUFF_H
