//----------------------------------------------------------------------------------------
/**
 * \file    pyramids.cpp
 * \author  Zuzana Slaninova
 * \using   asteroids
 * \date    2023
 * \brief   Pyramids World
 */
//----------------------------------------------------------------------------------------

#include <time.h>
#include <list>
#include <iostream>
#include "pgr.h"
#include "render_stuff.h"
#include "spline.h"
#include "data.h"

//curve data definition for extraDuck
const size_t  curveSize = 16U;

/// Control points of the animation curve for extraDuck
glm::vec3 curveData[] = {
  glm::vec3(-0.064, 0.088, -0.2),
  glm::vec3(-0.344, 0.178667, -0.2),
  glm::vec3(-0.792, 0.170667, -0.2),
  glm::vec3(-0.965333, -0.00800002, -0.2),
  glm::vec3(-0.965333, -0.285333, -0.2),
  glm::vec3(-0.746667, -0.562667, -0.2),
  glm::vec3(-0.464, -0.552, -0.2),
  glm::vec3(-0.133333, -0.442667, -0.2),
  glm::vec3(0.136, -0.304, -0.2),
  glm::vec3(0.205333, -0.133333, -0.2),
  glm::vec3(0.368, -0.0586667, -0.2),
  glm::vec3(0.757333, -0.0880001, -0.2),
  glm::vec3(0.850667, -0.186667, -0.2),
  glm::vec3(0.666667, -0.288, -0.2),
  glm::vec3(0.333333, -0.274667, -0.2),
  glm::vec3(0.136, -0.130667, -0.2)
};

//curve data definition for camera
const size_t cameraCurveSize = 4U;

/// Control points of the animation curve for camera
glm::vec3 cameraCurveData[] = {
    glm::vec3(-0.2f, 0.8f, -0.05f),
    glm::vec3(-0.2f, -0.8f, -0.05f),
    glm::vec3(0.2f, -0.8f, -0.05f),
    glm::vec3(0.2f, 0.8f, -0.05f),
};

glm::vec3 cameraCurvePosition = cameraCurveData[0];
glm::vec3 cameraCurveDirection;

extern SCommonShaderProgram shaderProgram;
extern bool useLighting;
extern bool fog;

typedef std::list<void *> GameObjectsList; 

struct GameState {

  int windowWidth;    // set by reshape callback
  int windowHeight;   // set by reshape callback

  float cameraElevationAngle; // in degrees = initially 0.0f

  bool snake;              // false;
  bool keyMap[KEYS_COUNT];    // false

  float elapsedTime;
  glm::mat4 viewMatrix;
  glm::mat4 projectionMatrix;

  int geo_index = 0; //index for flag geometry

  CameraEnum cameraMode;

  StencilIDs IDenum;

} gameState;

struct GameObjects {

  BalonObject *balon; // NULL

  GameObjectsList ducks;

  GameObjectsList pyramids_v1;
  GameObjectsList pyramids_v2;
  GameObjectsList pyramids_v3;

  GameObjectsList flags;

  GameObjectsList aloes;

  GameObjectsList sands;

  GameObjectsList feathers;

  GameObjectsList aloeAnims;

  Object* rock;

  DuckObject* extraDuck;

  Object* movingTexture;

  Object* carpet;

} gameObjects;

//
// moving with balon
//

void increaseBalonSpeed(float deltaSpeed = BALON_SPEED_INCREMENT) {

  gameObjects.balon->speed =
    std::min(gameObjects.balon->speed + deltaSpeed, BALON_SPEED_MAX);
}

void decreaseBalonSpeed(float deltaSpeed = BALON_SPEED_INCREMENT) {

  gameObjects.balon->speed =
    std::max(gameObjects.balon->speed - deltaSpeed, 0.0f);
}

void turnBalonLeft(float deltaAngle) {

  gameObjects.balon->viewAngle += deltaAngle;

  if(gameObjects.balon->viewAngle > 360.0f)
    gameObjects.balon->viewAngle -= 360.0f;

  float angle = glm::radians(gameObjects.balon->viewAngle);

  gameObjects.balon->direction.x = cos(angle);
  gameObjects.balon->direction.y = sin(angle);
}

void turnBalonRight(float deltaAngle) {

  gameObjects.balon->viewAngle -= deltaAngle;

  if(gameObjects.balon->viewAngle < 0.0f)
    gameObjects.balon->viewAngle += 360.0f;

  float angle = glm::radians(gameObjects.balon->viewAngle);

  gameObjects.balon->direction.x = cos(angle);
  gameObjects.balon->direction.y = sin(angle);
}

//
// end of moving with balon
//

//cleans objects
void cleanUpObjects(void) {

  // delete ducks
  while(!gameObjects.ducks.empty()) {
    delete gameObjects.ducks.back();
    gameObjects.ducks.pop_back();
  } 

  // delete sand
  while(!gameObjects.sands.empty()) {
    delete gameObjects.sands.back();
    gameObjects.sands.pop_back();
  } 

  // delete feather
  while (!gameObjects.feathers.empty()) {
      delete gameObjects.feathers.back();
      gameObjects.feathers.pop_back();
  }

  while (!gameObjects.aloeAnims.empty()) {
      delete gameObjects.aloeAnims.back();
      gameObjects.aloeAnims.pop_back();
  }

  while (!gameObjects.aloes.empty()) {
      delete gameObjects.aloes.back();
      gameObjects.aloes.pop_back();
  }

  while (!gameObjects.pyramids_v1.empty()) {
      delete gameObjects.pyramids_v1.back();
      gameObjects.pyramids_v1.pop_back();
  }

  while (!gameObjects.pyramids_v2.empty()) {
      delete gameObjects.pyramids_v2.back();
      gameObjects.pyramids_v2.pop_back();
  }

  while (!gameObjects.pyramids_v3.empty()) {
      delete gameObjects.pyramids_v3.back();
      gameObjects.pyramids_v3.pop_back();
  }

  while (!gameObjects.flags.empty()) {
      delete gameObjects.flags.back();
      gameObjects.flags.pop_back();
  }

  if (gameObjects.movingTexture != NULL) {
      delete gameObjects.movingTexture;
      gameObjects.movingTexture = NULL;
  }
}

// generates random position
glm::vec3 generateRandomPosition(void) {
 glm::vec3 newPosition;

    // position is generated randomly
    // coordinates are in range -1.0f ... 1.0f
    newPosition = glm::vec3(
      (float)(2.0 * (rand() / (double)RAND_MAX) - 1.0),
      (float)(2.0 * (rand() / (double)RAND_MAX) - 1.0),
      0.0f
    );

  return newPosition;
}

//creates duck
DuckObject* createDuck(void) {
 DuckObject* newDuck = new DuckObject;

  newDuck->destroyed = false;

  newDuck->startTime = gameState.elapsedTime;
  newDuck->currentTime = newDuck->startTime;

  newDuck->size = DUCK_SIZE;

  // generate motion direction
  newDuck->direction = glm::vec3(
    1.0f,
    0.0f,
    0.0f
  );
  newDuck->direction = glm::normalize(newDuck->direction);

  // position is generated randomly
  newDuck->position = generateRandomPosition();

  // motion speed 0.0f ... 1.0f
  newDuck->speed = DUCK_SPEED_MAX * (float)(rand() / (double)RAND_MAX) + 0.1f;

  newDuck->rotationSpeed = 0;

  return newDuck;
}

//creates extraDuck - duck on curve
DuckObject* createExtraDuck(void) {
    DuckObject* newDuck = new DuckObject;

    newDuck->destroyed = false;

    float elapsedTime = 0.001f * (float)glutGet(GLUT_ELAPSED_TIME);
    int timeMs = glutGet(GLUT_ELAPSED_TIME);
    float timePassed = timeMs * 0.001f - newDuck->startTime;
    newDuck->startTime += timePassed;
    float speed = newDuck->startTime * 0.9;
    newDuck->position = curveData[0];
    newDuck->direction = -glm::normalize(evaluateClosedCurve_1stDerivative(curveData, curveSize, speed));

    newDuck->startTime = gameState.elapsedTime;
    newDuck->currentTime = newDuck->startTime;

    newDuck->size = DUCK_SIZE;

    newDuck->rotationSpeed = 0;

    return newDuck;
}

//creates pyramid
Object* createPyramid(glm::vec3 pos, float size) {
    Object* newPyramid = new Object;

    newPyramid->destroyed = false;

    newPyramid->startTime = gameState.elapsedTime;
    newPyramid->currentTime = newPyramid->startTime;

    newPyramid->size = size;

   
    newPyramid->position = pos;
    
    newPyramid->speed = 0;
    

    newPyramid->direction = glm::vec3(
        0.5f,
        0.5f,
        0.5f
    );
    newPyramid->direction = glm::normalize(newPyramid->direction);

    return newPyramid;
}

//creates flag
Object* createFlag(glm::vec3 pos, float size) {
    Object* newFlag = new Object();

    newFlag->destroyed = false;

    newFlag->startTime = gameState.elapsedTime;
    newFlag->currentTime = newFlag->startTime;

    newFlag->size = size;


    newFlag->position = pos; 

    newFlag->speed = 0;

    newFlag->direction = glm::vec3(0.5f, 0.5f, 0.5f);
    newFlag->direction = glm::normalize(newFlag->direction);

    return newFlag;
}

//creates aloe
Object* createAloe(glm::vec3 pos, float size) {
    Object* newAloe = new Object;

    newAloe->destroyed = false;

    newAloe->startTime = gameState.elapsedTime;
    newAloe->currentTime = newAloe->startTime;

    newAloe->size = size;


    newAloe->position = pos;

    newAloe->speed = 0;

    newAloe->direction = glm::vec3(
        0.5f,
        0.5f,
        0.5f
    );
    newAloe->direction = glm::normalize(newAloe->direction);

    return newAloe;
}

//creates rock
Object* createRock(glm::vec3 pos, float size) {
    Object* newRock = new Object;

    newRock->destroyed = false;

    newRock->startTime = gameState.elapsedTime;
    newRock->currentTime = newRock->startTime;

    newRock->size = size; 


    newRock->position = pos; 

    newRock->speed = 0;

    newRock->direction = glm::vec3(
        0.5f,
        0.5f,
        0.5f
    );
    newRock->direction = glm::normalize(newRock->direction);

    return newRock;
}

//creates moving texture
Object* createTexture(void) {
    Object* newTexture1 = new Object;
    newTexture1->size = 0.45f;
    newTexture1->position = glm::vec3(0.05f, -0.4f, -1.95f);
    newTexture1->direction = glm::vec3(0.0f, 1.0f, 1.0f);
    newTexture1->speed = 0.0f;

    newTexture1->destroyed = false;

    newTexture1->startTime = gameState.elapsedTime;
    newTexture1->currentTime = newTexture1->startTime;

    return newTexture1;
}

//creates carpet
Object* createCarpet(glm::vec3 pos, float size) {
    Object* newCarpet = new Object;

    newCarpet->destroyed = false;

    newCarpet->startTime = gameState.elapsedTime;
    newCarpet->currentTime = newCarpet->startTime;

    newCarpet->size = size; // ROCK_SIZE;


    newCarpet->position = pos; // glm::vec3(0.8f, 0.99f, -1.4f);

    newCarpet->speed = 0;

    newCarpet->direction = glm::vec3(
        0.5f,
        0.5f,
        0.5f
    );
    newCarpet->direction = glm::normalize(newCarpet->direction);

    return newCarpet;
}

//creates sand animation
void insertSand(const glm::vec3& position) {

    SandObject* newSand = new SandObject;

    newSand->speed = 0.0f;
    newSand->destroyed = false;

    newSand->startTime = gameState.elapsedTime;
    newSand->currentTime = newSand->startTime;

    newSand->size = BILLBOARD_SIZE;
    newSand->direction = glm::vec3(0.0f, 0.0f, 1.0f);

    newSand->frameDuration = 0.1f;
    newSand->textureFrames = 16;

    newSand->position = position;

    gameObjects.sands.push_back(newSand);
}

//creates feather animation
void insertFeather(const glm::vec3& position) {

    SandObject* newFeather = new SandObject;

    newFeather->speed = 0.0f;
    newFeather->destroyed = false;

    newFeather->startTime = gameState.elapsedTime;
    newFeather->currentTime = newFeather->startTime;

    newFeather->size = FEATHER_BILLBOARD_SIZE;
    newFeather->direction = glm::vec3(0.0f, 0.0f, 1.0f);

    newFeather->frameDuration = 0.1f;
    newFeather->textureFrames = 16;

    newFeather->position = position;

    gameObjects.feathers.push_back(newFeather);
}

//creates aloe animation
void insertAloeAnim(const glm::vec3& position) {

    SandObject* newAloeA = new SandObject;

    newAloeA->speed = 0.0f;
    newAloeA->destroyed = false;

    newAloeA->startTime = gameState.elapsedTime;
    newAloeA->currentTime = newAloeA->startTime;

    newAloeA->size = BILLBOARD_SIZE;
    newAloeA->direction = glm::vec3(0.0f, 0.0f, 1.0f);

    newAloeA->frameDuration = 0.1f;
    newAloeA->textureFrames = 16;

    newAloeA->position = position;

    gameObjects.aloeAnims.push_back(newAloeA);
}

//collision check
bool spheresIntersection(const glm::vec3& center1, float radius1, const glm::vec3& center2, float radius2) {
    if (distance(center1, center2) <= (radius1 + radius1))
    {
        return true;
    }
    else
    {
        return false;
    }
}

//check collosion between duck and balon -> insert feather animation
void checkCollisions(void) {
    GameObjectsList::iterator it;

    for (it = gameObjects.ducks.begin(); it != gameObjects.ducks.end(); ++it) {
        DuckObject* duck = (DuckObject*)(*it);

        if (duck->destroyed == false) {
            if (spheresIntersection(gameObjects.balon->position, gameObjects.balon->size, duck->position, duck->size) == true) {
                duck->destroyed = true;	                       // mark asteroid dead
                insertFeather(gameObjects.balon->position);  // insert explostion billboard
            }
        }
    }
}

void restartGame(void) {

  cleanUpObjects();

  gameState.cameraMode = TOP;
  glutPassiveMotionFunc(NULL);
  fog = false;

  gameState.elapsedTime = 0.001f * (float)glutGet(GLUT_ELAPSED_TIME); // milliseconds => seconds

  // initialize space ship
  if(gameObjects.balon == NULL)
    gameObjects.balon = new BalonObject;

  gameObjects.balon->position = glm::vec3(0.0f, 0.0f, 0.0f);
  gameObjects.balon->viewAngle = 90.0f; // degrees
  gameObjects.balon->direction = glm::vec3(cos(glm::radians(gameObjects.balon->viewAngle)), sin(glm::radians(gameObjects.balon->viewAngle)), 0.0f);
  gameObjects.balon->speed = 0.0f;
  gameObjects.balon->size = BALON_SIZE;
  gameObjects.balon->destroyed = false;
  gameObjects.balon->startTime = gameState.elapsedTime;
  gameObjects.balon->currentTime = gameObjects.balon->startTime;

  // initialize ducks
  for(int i=0; i<DUCKS_COUNT_MIN; i++) {
    DuckObject* newDuck = createDuck();
    
    gameObjects.ducks.push_back(newDuck); 
  }

  Object* newPyramid;
  newPyramid = createPyramid(glm::vec3(1.0f, 0.5f, -1.0f), 0.6f);
  gameObjects.pyramids_v1.push_back(newPyramid);
  newPyramid = createPyramid(glm::vec3(0.7f, -0.6f, -1.2f), 0.3f);
  gameObjects.pyramids_v1.push_back(newPyramid);

  newPyramid = createPyramid(glm::vec3(-0.6f, 0.4f, -1.0f), 0.2f);
  gameObjects.pyramids_v2.push_back(newPyramid);
  newPyramid = createPyramid(glm::vec3(-0.6f, -0.2f, -1.0f), 0.3f);
  gameObjects.pyramids_v2.push_back(newPyramid);
  newPyramid = createPyramid(glm::vec3(-0.6f, -0.8f, -1.0f), 0.2f);
  gameObjects.pyramids_v2.push_back(newPyramid);
  
  newPyramid = createPyramid(glm::vec3(-0.6f, 0.7f, -1.0f), 0.2f);
  gameObjects.pyramids_v3.push_back(newPyramid);
  newPyramid = createPyramid(glm::vec3(0.55f, -0.2f, -1.3f), 0.15f);
  gameObjects.pyramids_v3.push_back(newPyramid);
  newPyramid = createPyramid(glm::vec3(-0.6f, 0.85f, -1.0f), 0.1f);
  gameObjects.pyramids_v3.push_back(newPyramid);

  Object* newFlag;
  newFlag = createFlag(glm::vec3(1.0f, 0.5f, -0.4f), 0.1f);
  gameObjects.flags.push_back(newFlag);
  newFlag = createFlag(glm::vec3(-0.57f, -0.2f, -0.7f), 0.1f);
  gameObjects.flags.push_back(newFlag);

  Object* newAloe;
  newAloe = createAloe(glm::vec3(-0.3f, 0.5f, -1.0f), 0.05f);
  gameObjects.aloes.push_back(newAloe);
  newAloe = createAloe(glm::vec3(-0.3f, 0.4f, -1.0f), 0.05f);
  gameObjects.aloes.push_back(newAloe);
  newAloe = createAloe(glm::vec3(-0.15f, 0.0f, -1.0f), 0.08f);
  gameObjects.aloes.push_back(newAloe);
  
  gameState.cameraElevationAngle = 0.0f;

  gameObjects.extraDuck = createExtraDuck();

  gameObjects.rock = createRock(glm::vec3(-0.2f, -0.8f, -1.1f), 0.06f);

  gameObjects.carpet = createCarpet(glm::vec3(-0.2f, 0.6f, -0.3f), 0.18f);

  // reset key map
  for(int i=0; i<KEYS_COUNT; i++)
    gameState.keyMap[i] = false;

  Object* newTexture;
  newTexture = createTexture();

  gameState.snake = false;
}

//draw window content in specific camera mode
void drawWindowContents() {
    if (gameState.cameraMode == TOP) {
        // setup parallel projection
        glm::mat4 orthoProjectionMatrix = glm::ortho(
            -SCENE_WIDTH, SCENE_WIDTH,
            -SCENE_HEIGHT, SCENE_HEIGHT,
            -10.0f * SCENE_DEPTH, 10.0f * SCENE_DEPTH
        );
        // static viewpoint - top view
        glm::mat4 orthoViewMatrix = glm::lookAt(
            glm::vec3(0.0f, 0.0f, 1.0f),
            glm::vec3(0.0f, 0.0f, 0.0f),
            glm::vec3(0.0f, 1.0f, 0.0f)
        );

        // setup camera & projection transform
        gameState.viewMatrix = orthoViewMatrix;
        gameState.projectionMatrix = orthoProjectionMatrix;
    }

    if(gameState.cameraMode == FREE) {   

        // compute correctly camera position and up vector using balon parameters
        glm::vec3 cameraPosition = gameObjects.balon->position;
        glm::vec3 cameraUpVector = glm::vec3(0.0f, 0.0f, 1.0f);

        glm::vec3 cameraViewDirection = gameObjects.balon->direction;

        // update camera view direction vector and cameraUpVector by gameState.cameraElevationAngle variable
        glm::vec3 rotationAxis = glm::cross(cameraViewDirection, cameraUpVector);
        glm::mat4 cameraTransform = glm::rotate(glm::mat4(1.0f), glm::radians(gameState.cameraElevationAngle), rotationAxis);

        cameraUpVector = glm::vec3(cameraTransform * glm::vec4(cameraUpVector, 0.0f));
        cameraViewDirection = glm::vec3(cameraTransform * glm::vec4(cameraViewDirection, 0.0f));

        glm::vec3 cameraCenter = cameraPosition + cameraViewDirection;

        gameState.viewMatrix = glm::lookAt(
            cameraPosition,
            cameraCenter,
            cameraUpVector
        );

        gameState.projectionMatrix = glm::perspective(glm::radians(60.0f), gameState.windowWidth/(float)gameState.windowHeight, 0.1f, 10.0f);
    }

    if (gameState.cameraMode == SIDE)
    {
        //gameObjects.balon->position = glm::vec3(-0.5f, -0.5f, 0.0f);

        //glm::vec3 cameraPosition = gameObjects.balon->position;

        gameState.viewMatrix = glm::lookAt(
            glm::vec3(-0.5f, -0.5f, 0.0f), //cameraPosition, //position
            glm::vec3(0.0f, 0.0f, -0.3f), //target
            glm::vec3(1.0f, 1.0f, 0.0f) //up
        );

        gameState.projectionMatrix = glm::ortho(
            -SCENE_WIDTH, SCENE_WIDTH,
            -SCENE_HEIGHT, SCENE_HEIGHT,
            -10.0f * SCENE_DEPTH, 10.0f * SCENE_DEPTH
        );

        
    }

    if (gameState.cameraMode == CURVE) {

        // compute correctly camera position and up vector using space ship parameters
        glm::vec3 cameraPosition = cameraCurvePosition;
        glm::vec3 cameraUpVector = glm::vec3(0.0f, 0.0f, 1.0f);

        //glm::vec3 cameraViewDirection = cameraCurveDirection; //glm::vec3(0.0f, 0.0f, -0.3f);

        //glm::vec3 cameraCenter = cameraPosition + cameraViewDirection;

        gameState.viewMatrix = glm::lookAt(
            cameraPosition,
            glm::vec3(0.0f),
            cameraUpVector
        );

        gameState.projectionMatrix = glm::perspective(glm::radians(60.0f), gameState.windowWidth / (float)gameState.windowHeight, 0.1f, 10.0f);
    }

    glUseProgram(shaderProgram.program);
    glUniform1f(shaderProgram.timeLocation, gameState.elapsedTime);

    glUniform3fv(shaderProgram.reflectorPositionLocation, 1, glm::value_ptr(gameObjects.balon->position));
    glUniform3fv(shaderProgram.reflectorDirectionLocation, 1, glm::value_ptr(gameObjects.balon->direction));
    glUseProgram(0);

    glDepthFunc(GL_LEQUAL);
    // draw skybox
    drawSkybox(gameState.viewMatrix, gameState.projectionMatrix);
    glDepthFunc(GL_LESS);

    glEnable(GL_STENCIL_TEST);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    // draw space ship
    drawBalon(gameObjects.balon, gameState.viewMatrix, gameState.projectionMatrix);

    // draw ducks
    for(GameObjectsList::iterator it = gameObjects.ducks.begin(); it != gameObjects.ducks.end(); ++it) {
        CHECK_GL_ERROR(); 
        DuckObject* duck = (DuckObject*)(*it);
        drawDuck(duck, gameState.viewMatrix, gameState.projectionMatrix);
    }

    glStencilFunc(GL_ALWAYS, PYRAMID, -1);
    for (GameObjectsList::iterator it = gameObjects.pyramids_v1.begin(); it != gameObjects.pyramids_v1.end(); ++it)
    {
        Object* p = (Object*)(*it);
        drawPyramid(p, gameState.viewMatrix, gameState.projectionMatrix);
    }
    glStencilFunc(GL_ALWAYS, ZERO, -1);
    for (GameObjectsList::iterator it = gameObjects.pyramids_v2.begin(); it != gameObjects.pyramids_v2.end(); ++it)
    {
        Object* p = (Object*)(*it);
        drawPyramid2(p, gameState.viewMatrix, gameState.projectionMatrix);
    }

    for (GameObjectsList::iterator it = gameObjects.pyramids_v3.begin(); it != gameObjects.pyramids_v3.end(); ++it)
    {
        Object* p = (Object*)(*it);
        drawPyramid3(p, gameState.viewMatrix, gameState.projectionMatrix);
    }

    for (GameObjectsList::iterator it = gameObjects.flags.begin(); it != gameObjects.flags.end(); ++it) {
        CHECK_GL_ERROR();

        Object* flag = (DuckObject*)(*it);
        drawFlag(flag, gameState.viewMatrix, gameState.projectionMatrix, gameState.geo_index);
    }
    glStencilFunc(GL_ALWAYS, ALOE, -1);
    for (GameObjectsList::iterator it = gameObjects.aloes.begin(); it != gameObjects.aloes.end(); ++it)
    {
        Object* aloe = (Object*)(*it);
        drawAloe(aloe, gameState.viewMatrix, gameState.projectionMatrix);
    }
    glStencilFunc(GL_ALWAYS, ZERO, -1);
    //draw extra duck
    drawExtraDuck(gameObjects.extraDuck, gameState.viewMatrix, gameState.projectionMatrix);

    // draw moving texture
    //drawTexture1(gameState.viewMatrix, gameState.projectionMatrix);
  
    // draw sands with depth test disabled
    glDisable(GL_DEPTH_TEST);

    for(GameObjectsList::iterator it = gameObjects.sands.begin(); it != gameObjects.sands.end(); ++it) {
        SandObject* sand = (SandObject *)(*it);
        drawSand(sand, gameState.viewMatrix, gameState.projectionMatrix); 
    }
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_DEPTH_TEST);
    for (GameObjectsList::iterator it = gameObjects.feathers.begin(); it != gameObjects.feathers.end(); ++it) {
        SandObject* feather = (SandObject*)(*it);
        drawFeather(feather, gameState.viewMatrix, gameState.projectionMatrix);
    }
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_DEPTH_TEST);
    for (GameObjectsList::iterator it = gameObjects.aloeAnims.begin(); it != gameObjects.aloeAnims.end(); ++it) {
        SandObject* aloeA = (SandObject*)(*it);
        drawAloeA(aloeA, gameState.viewMatrix, gameState.projectionMatrix);
    }
    glEnable(GL_DEPTH_TEST);

    glStencilFunc(GL_ALWAYS, ROCK, -1);
    if(gameObjects.rock != NULL && !gameObjects.rock->destroyed) drawRock(gameObjects.rock, gameState.viewMatrix, gameState.projectionMatrix);

    glStencilFunc(GL_ALWAYS, ZERO, -1);
    if (gameState.snake == true) {
        if (gameObjects.movingTexture != NULL)
            drawTexture(gameObjects.movingTexture, gameState.viewMatrix, gameState.projectionMatrix);
    }

    drawCarpet(gameObjects.carpet, gameState.viewMatrix, gameState.projectionMatrix);

    glDisable(GL_STENCIL_TEST);
}

// Called to update the display. You should call glutSwapBuffers after all of your
// rendering to display what you rendered.
void displayCallback() {

    glEnable(GL_DEPTH_TEST);
    GLbitfield mask = GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT;
    glClear(mask);

    drawWindowContents();

    glutSwapBuffers();
}

// Called whenever the window is resized. The new window size is given, in pixels.
// This is an opportunity to call glViewport or glScissor to keep up with the change in size.
void reshapeCallback(int newWidth, int newHeight) {

  gameState.windowWidth = newWidth;
  gameState.windowHeight = newHeight;

  glViewport(0, 0, (GLsizei) newWidth, (GLsizei) newHeight);
}

void updateObjects(float elapsedTime) {

  // update balon 
  float timeDelta = elapsedTime - gameObjects.balon->currentTime;
  gameObjects.balon->currentTime = elapsedTime;
  gameObjects.balon->position += timeDelta * gameObjects.balon->speed * gameObjects.balon->direction;

  // update ducks
  GameObjectsList::iterator it = gameObjects.ducks.begin();
  while(it != gameObjects.ducks.end()) {
    DuckObject* duck = (DuckObject*)(*it);

    if(duck->destroyed == true) {
      it = gameObjects.ducks.erase(it);
    }
    else {
      // update duck
      float timeDelta = elapsedTime - duck->currentTime;

      duck->currentTime = elapsedTime;
      duck->position += timeDelta * duck->speed * duck->direction;

      ++it;
    }
  }
  
  it = gameObjects.flags.begin();
  while (it != gameObjects.flags.end()) {
      Object* flag = (Object*)(*it);

      flag->currentTime = elapsedTime;
      drawFlag(flag, gameState.viewMatrix, gameState.projectionMatrix, gameState.geo_index);
      ++it;
  }

  // update sand billboards
  it = gameObjects.sands.begin();
  while (it != gameObjects.sands.end()) {
      SandObject* sand = (SandObject*)(*it);

      // update sand
      sand->currentTime = elapsedTime;

      if (sand->currentTime > sand->startTime + sand->textureFrames * sand->frameDuration)
          sand->destroyed = true;

      if (sand->destroyed == true) {
          it = gameObjects.sands.erase(it);
      }
      else {
          ++it;
      }
  }

  //udate feather billboards
  it = gameObjects.feathers.begin();
  while (it != gameObjects.feathers.end()) {
      SandObject* feather = (SandObject*)(*it);

      // update sand
      feather->currentTime = elapsedTime;

      if (feather->currentTime > feather->startTime + feather->textureFrames * feather->frameDuration)
          feather->destroyed = true;

      if (feather->destroyed == true) {
          it = gameObjects.feathers.erase(it);
      }
      else {
          ++it;
      }
  }

  //udate aloe billboards
  it = gameObjects.aloeAnims.begin();
  while (it != gameObjects.aloeAnims.end()) {
      SandObject* aloeA = (SandObject*)(*it);

      // update sand
      aloeA->currentTime = elapsedTime;

      if (aloeA->currentTime > aloeA->startTime + aloeA->textureFrames * aloeA->frameDuration)
          aloeA->destroyed = true;

      if (aloeA->destroyed == true) {
          it = gameObjects.aloeAnims.erase(it);
      }
      else {
          ++it;
      }
  }

  if (gameObjects.rock != NULL && gameObjects.rock->destroyed == true) {
      delete gameObjects.rock;
      gameObjects.rock = NULL;
  }

  gameObjects.carpet->currentTime = elapsedTime;

  //updateTexture1(elapsedTime);
}

// Callback responsible for the scene update
void timerCallback(int) {

  // update scene time
  gameState.elapsedTime = 0.001f * (float)glutGet(GLUT_ELAPSED_TIME); // milliseconds => seconds

  gameState.geo_index = (gameState.geo_index + 1) % 5;

  //extra duck curve
  int timeMs = glutGet(GLUT_ELAPSED_TIME);
  float timePassed = timeMs * 0.001f - gameObjects.extraDuck->startTime;
  gameObjects.extraDuck->startTime += timePassed;
  float speed = gameObjects.extraDuck->startTime * 0.9;
  gameObjects.extraDuck->position = evaluateClosedCurve(curveData, curveSize, speed);
  gameObjects.extraDuck->direction = -glm::normalize(evaluateClosedCurve_1stDerivative(curveData, curveSize, speed));

  //camera curve
  if (gameState.cameraMode == CURVE) {
      cameraCurvePosition = evaluateClosedCurve(cameraCurveData, cameraCurveSize, speed);
      cameraCurveDirection = -glm::normalize(evaluateClosedCurve_1stDerivative(cameraCurveData, cameraCurveSize, speed));
  }

  if(gameState.keyMap[KEY_RIGHT_ARROW] == true)
    turnBalonRight(BALON_VIEW_ANGLE_DELTA);

  if(gameState.keyMap[KEY_LEFT_ARROW] == true)
    turnBalonLeft(BALON_VIEW_ANGLE_DELTA);

  if(gameState.keyMap[KEY_UP_ARROW] == true)
    increaseBalonSpeed();

  if(gameState.keyMap[KEY_DOWN_ARROW] == true)
    decreaseBalonSpeed();

  if ((gameState.snake == true) && (gameObjects.movingTexture != NULL)) {
      gameObjects.movingTexture->currentTime = gameState.elapsedTime;
  }

  checkCollisions();

  // update objects in the scene
  updateObjects(gameState.elapsedTime);

  // generate new ducks randomly
  if(gameObjects.ducks.size() < DUCKS_COUNT_MIN) {
    int howManyDucks = rand() % (DUCKS_COUNT_MAX - DUCKS_COUNT_MIN + 1);

    for(int i=0; i<howManyDucks; i++) {
      DuckObject* newDuck = createDuck();

      gameObjects.ducks.push_back(newDuck); 
    }
  }
  if (gameState.snake == true) {
      gameState.keyMap[KEY_SPACE] = false;
      if (gameObjects.movingTexture == NULL) {
          // if game over and banner still not created -> create banner
          gameObjects.movingTexture = createTexture();
      }
  }

  // set timeCallback next invocation
  glutTimerFunc(33, timerCallback, 0);

  glutPostRedisplay();
}

void mouseCb(int buttonPressed, int buttonState, int mouseX, int mouseY) {
    if ((buttonPressed == GLUT_LEFT_BUTTON) && (buttonState == GLUT_DOWN) && gameState.cameraMode == TOP) {
        int objectID = 0;
        glReadPixels(mouseX, gameState.windowHeight - mouseY - 1, 1, 1, GL_STENCIL_INDEX, GL_UNSIGNED_BYTE, &objectID);
        float x = (float)mouseX / gameState.windowWidth * 2 -1;
        float y = (float)mouseY / gameState.windowWidth * 2 - 1;
        //std::cout << "(" << x << ", " << -y << ", -0.2)," << std::endl;
        if (objectID == PYRAMID) insertSand(glm::vec3(x, -y, 0.0f));
        if (objectID == ROCK) gameObjects.rock->destroyed = true;
        if (objectID == ALOE) insertAloeAnim(glm::vec3(x, -y, 0.0f));
    }
}

// Called when mouse is moving while no mouse buttons are pressed.
void passiveMouseMotionCallback(int mouseX, int mouseY) {

  if(mouseY != gameState.windowHeight/2) {

    float cameraElevationAngleDelta = 0.5f * (mouseY - gameState.windowHeight/2);

    if(fabs(gameState.cameraElevationAngle+cameraElevationAngleDelta) < CAMERA_ELEVATION_MAX)
      gameState.cameraElevationAngle += cameraElevationAngleDelta;

      // set mouse pointer to the window center
      glutWarpPointer(gameState.windowWidth/2, gameState.windowHeight/2);

      glutPostRedisplay();
    }
}

// Called whenever a key on the keyboard was pressed. The key is given by the "keyPressed"
// parameter, which is in ASCII. It's often a good idea to have the escape key (ASCII value 27)
// to call glutLeaveMainLoop() to exit the program.
void keyboardCallback(unsigned char keyPressed, int mouseX, int mouseY) {
  
  switch(keyPressed) {
    case 27: // escape
#ifndef __APPLE__
        glutLeaveMainLoop();
#else
        exit(0);
#endif
      break;
    case 'r': // restart game
        restartGame();
        break;

    //camera modes
    case 's': //switch to side static camera
        gameState.cameraMode = SIDE;
        glutPassiveMotionFunc(NULL);
        break;
    case 'c': // switch to free camera
        if (gameState.cameraMode == SIDE)
        {
            gameObjects.balon->position = glm::vec3(-0.5f, -0.5f, 0.0f); //side camera position
        }
        gameState.cameraMode = FREE;
        glutPassiveMotionFunc(passiveMouseMotionCallback);
        glutWarpPointer(gameState.windowWidth/2, gameState.windowHeight/2);
        break;
    case 'h': //switch to curve camera
        gameState.cameraMode = CURVE;
        glutPassiveMotionFunc(NULL);
        break;
    case 't': //switch to top static camera
        gameState.cameraMode = TOP;
        glutPassiveMotionFunc(NULL);
        break;

    case 'f': //switch fog
        fog = !fog;
        break;
    case 'm': //switch snake
        gameState.snake = !gameState.snake;
        break;
      ;
  }
}

// The special keyboard callback is triggered when keyboard function or directional
// keys are pressed.
void specialKeyboardCallback(int specKeyPressed, int mouseX, int mouseY) {


  switch (specKeyPressed) {
    case GLUT_KEY_RIGHT:
      gameState.keyMap[KEY_RIGHT_ARROW] = true;
      break;
    case GLUT_KEY_LEFT:
      gameState.keyMap[KEY_LEFT_ARROW] = true;
      break;
    case GLUT_KEY_UP:
      gameState.keyMap[KEY_UP_ARROW] = true;
      break;
    case GLUT_KEY_DOWN:
      gameState.keyMap[KEY_DOWN_ARROW] = true;
      break;
    default:
      ; // printf("Unrecognized special key pressed\n");
  }
}

// The special keyboard callback is triggered when keyboard function or directional
// keys are released.
void specialKeyboardUpCallback(int specKeyReleased, int mouseX, int mouseY) {

  switch (specKeyReleased) {
	case GLUT_KEY_RIGHT:
      gameState.keyMap[KEY_RIGHT_ARROW] = false;
      break;
    case GLUT_KEY_LEFT:
      gameState.keyMap[KEY_LEFT_ARROW] = false;
      break;
    case GLUT_KEY_UP:
      gameState.keyMap[KEY_UP_ARROW] = false;
      break;
    case GLUT_KEY_DOWN:
      gameState.keyMap[KEY_DOWN_ARROW] = false;
      break;
		default:
      ; // printf("Unrecognized special key released\n");
  }
}

// Called after the window and OpenGL are initialized. Called exactly once, before the main loop.
void initializeApplication() {

  // initialize random seed
  srand ((unsigned int)time(NULL));

  // initialize OpenGL
  glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
  glEnable(GL_DEPTH_TEST);

  useLighting = true;

  gameState.cameraMode = TOP;

  gameState.snake = false;

  // initialize shaders
  initializeShaderPrograms();
  // create geometry for all models used
  initializeModels();

  // initialize shaders and geometry for Texture1
  //initTexture1();

  gameObjects.balon = NULL;
  gameObjects.movingTexture = NULL;

  restartGame();
}

void finalizeApplication(void) {

  cleanUpObjects();

  delete gameObjects.balon;
  gameObjects.balon = NULL;

  // delete models
  cleanupModels();

  // delete shaders
  cleanupShaderPrograms();

  //cleanupTexture1();
}

int main(int argc, char** argv) {

  // initialize windowing system
  glutInit(&argc, argv);

#ifndef __APPLE__
  glutInitContextVersion(pgr::OGL_VER_MAJOR, pgr::OGL_VER_MINOR);
  glutInitContextFlags(GLUT_FORWARD_COMPATIBLE);

  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
#else
  glutInitDisplayMode(GLUT_3_2_CORE_PROFILE | GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
#endif

  // initial window size
  glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
  glutCreateWindow(WINDOW_TITLE);

  glutDisplayFunc(displayCallback);
  // register callback for change of window size
  glutReshapeFunc(reshapeCallback);
  // register callbacks for keyboard
  glutKeyboardFunc(keyboardCallback);
  //glutKeyboardUpFunc(keyboardUpCallback);
  glutSpecialFunc(specialKeyboardCallback);     // key pressed
  glutSpecialUpFunc(specialKeyboardUpCallback); // key released
  glutMouseFunc(mouseCb);

  glutTimerFunc(33, timerCallback, 0);

  // initialize PGR framework (GL, DevIl, etc.)
  if(!pgr::initialize(pgr::OGL_VER_MAJOR, pgr::OGL_VER_MINOR))
    pgr::dieWithError("pgr init failed, required OpenGL not supported?");

  initializeApplication();

#ifndef __APPLE__
  glutCloseFunc(finalizeApplication);
#else
  glutWMCloseFunc(finalizeApplication);
#endif

  glutMainLoop();

  return 0;
}
